# -*- coding: utf-8 -*-
"""
Module containing functions for in-memory and in-database table creation and
in-database table dropping.

Supports:
* table schema/accounts/owners/namespaces. SQLAlchemy calls them "schema."
* prefixed table names for standard table prototypes
* loading arbitrary tables from existing database
* creating standard prototype tables

Supported actions are:
* creating in-memory canonical tables

Glossary
--------
* table : canonical table name string, like "site" or "origin"
* tablename : the fully-qualified table name string (includes any owner, prefix)
* mapped table : a SQLAlchemy ORM table class
* table instance / row : an instance of a mapped table

"""
# TODO: support plugging in additional schema
import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base

from pisces.schema.util import PiscesMeta
import pisces.tables.css3 as css3
import pisces.tables.kbcore as kb

def make_table_names(*tables, **kwargs):
    """
    Get table name strings in the format: {owner.}{prefix}tablename

    This function does no checking that supplied table names are in the
    indicated schema.

    Parameters
    ----------
    tables : str
        Desired table names.  If omitted, a schema must be specified, and all
        core tables for the schema are used.
    schema : str  {'css3', 'kbcore'}
        Which set of core tables are being used. If omitted, one or more tables
        must be specified.
    prefix : str
        All table names will have the provided prefix.
        e.g. "TA_" for "TA_origin", "TA_wfdisc", etc.
    owner : str
        e.g. "myuser" for "myuser.origin", "myuser.wfdisc", etc.

    Returns
    -------
    dict
        The desired formatted table name strings, keyed by canonical name.

    Raises
    ------
    ValueError : unknown schema specified, or tablenames omitted and schema
                 not specified

    Examples
    --------
    >>> make_table_names('site', 'origin', owner='global', prefix='TA_')
    {'site': 'global.TA_site', 'origin': 'global.TA_origin'}

    >>> make_table_names(schema='css3', prefix='TA_')
    {'affiliation': 'TA_affiliation', 'assoc': 'TA_assoc', ...}


    """
    # TODO: is this function really necessary?  remove it?
    # >>> tables = [owner + '.' + prefix + table.name for table in coretables]
    # >>> tables = [prefix + table.name for table in coretables]
    prefix = kwargs.get('prefix', '')
    owner = kwargs.get('owner', None)
    schema = kwargs.get('schema', None)

    if not tables:
        if schema == 'css3':
            CORETABLES = css3.CORETABLES
        elif schema == 'kbcore':
            CORETABLES = kb.CORETABLES
        else:
            msg = "Unknown schema: {}".format(schema)
            raise ValueError(msg)

        tables = CORETABLES.keys()

    if owner:
        fmt = "{owner}.{prefix}{table}"
    else:
        fmt = "{prefix}{table}"

    tablenames = {}
    for table in tables:
        tablenames[table] = fmt.format(owner=owner, prefix=prefix, table=table)

    return tablenames


def split_table_names(*tablenames, **kwargs):
    """
    Splits full table names into a (owner, prefix, tablename) 3-tuples.

    Parameters
    ----------
    tablenames : str
    schema : str {'css3', 'kbcore'} Default 'css3'.
    split_prefix : bool, default True
        If True, return (owner, prefix, tablename) tuples,
        otherwise return (owner, tablename) tuples.

    Returns
    -------
    list
        Corresponding list of (owner, tablename) or (owner, prefix, tablename)
        tuples strings. Empty owner or prefix are ''.

    Examples
    --------
    >>> tablenames = ['global.site', 'global.sitechan', 'TA_site',
    ...               'different_acct.my_origin', 'myfriend.discrim_last']
    >>> split_table_names(*tablenames)
    [('global', '', 'site'), ('global', '', 'sitechan'), ('', 'TA_', 'site'),
    ('different_acct', 'my_', 'origin'), ('myfriend', '', 'discrim_last')]

    """
    split_prefix = kwargs.pop('split_prefix', True)
    schema = kwargs.get('schema', 'css3')
    if schema == 'css3':
        schematables = css3.CORETABLES.keys()
    elif schema == 'kbcore':
        schematables = kb.CORETABLES.keys()
    else:
        msg = "Unknown schema: {}".format(schema)
        raise ValueError(msg)

    out = []
    for fulltablename in tablenames:
        try:
            owner, prefixed_tablename = fulltablename.split('.')
        except ValueError:
            owner = ''
            prefixed_tablename = fulltablename

        if split_prefix:
            # if a table 'endswith' a table in the schema, the part before is the
            # table prefix
            for schematable in schematables:
                head, sep, tail = prefixed_tablename.rpartition(schematable)
                if sep and not tail:
                    prefix = head
                    tablename = schematable
                elif not sep:
                    prefix = ''
                    tablename = tail

            out.append((owner, prefix, tablename))
        else:
            out.append((owner, prefixed_tablename))


    return out


def make_tables(*tables, **kwargs):
    """
    Create mapped SQLAlchemy ORM table classes.

    If no owner or prefix is indicated, the returned tables are the same as
    those in pisces.tables.<schema> .  The user is encouraged to import these
    classes directly. If an owner is indicated, returned tables inhererit from
    the same SQLA declarative base.

    Parameters
    ----------
    tables : str
        Desired table names.  Table name must be known to the schema. 
        If omitted, a schema must be specified, and all core tables for the
        schema are returned.
    schema : str  ("css3" or "kbcore")
        Which set of core tables are being used.  Required.
    prefix : str
        All table names will have the provided prefix.
        e.g. "TA_" for "TA_origin", "TA_wfdisc", etc.
    owner : str
        e.g. "myuser" for "myuser.origin", "myuser.wfdisc", etc.

    Returns
    -------
    dict
        Corresponding dict of mapped SQLAlchemy table classes, keyed by
        canonical name.

    Examples
    --------
    >>> # makes a dict of CSS3.0 'TA_site' and 'TA_sitechan' mapped classes
    >>> tables = make_tables('site', 'sitechan', schema='css3', prefix='TA_')

    >>> # make a dict of all standard kbcore tables
    >>> tables = make_tables(schema='kbcore')

    >>> # make a dict of some tables from the 'myowner' account
    >>> tables = make_tables('origin', 'event', schema='kbcore', owner='myowner')

    """
    # if there is an owner or a prefix, the tables need to be created
    # otherwise, just return the existing ones.

    # TODO: find a way to not repeat arguments with all these functions
    prefix = kwargs.get('prefix', '')
    owner = kwargs.get('owner', None)
    schema = kwargs.get('schema', 'css3')

    if schema == 'css3':
        CORETABLES = css3.CORETABLES
    elif schema == 'kbcore':
        CORETABLES = kb.CORETABLES
    else:
        msg = "Unknown schema: {}".format(schema)
        raise ValueError(msg)

    non_schema_tables = set(tables) - set(CORETABLES.keys())
    if non_schema_tables:
        msg = "Unknown tables: {}".format(non_schema_tables)
        raise ValueError(msg)

    # I don't use 'owner' here, b/c we will be making classes. 'owner' is used
    # in making the declarative base class and a non-owner-qualified tablename
    # is used in the __tablename__
    tablenames = make_table_names(*tables, schema=schema, prefix=prefix)

    if owner:
        parents = (declarative_base(metadata=sa.metaData(schema=owner)),)
    else:
        parents = ()

    # "out" will hold concrete classes
    # go through one by one & repace names w/ classes
    # if a standard table, get the prototype concrete class
    out = {}
    for table, tablename in tablenames.items():
        prototype = CORETABLES[table].prototype
        # TODO: capitalize tablename?
        out[table] = type(tablename, parents + (prototype,),
                          {'__tablename__': tablename})

    return out


def load_tables(session, *tables, **kwargs):
    """
    Load existing database tables into mapped SQLAlchemy table classes.

    This function is not recommended.  Instead, please subclass from an
    existing abstract table in the schema, or use the SQLAlchemy ORM to create
    a class for your table.

    session : sqlalchemy.orm.Session
        Session connected to the target database.
    tables : str
        Desired table names.  Names must be part of the schema. If omitted, a
        schema must be specified, and all core tables for the schema are
        returned.
    schema : str  {'css3', 'kbcore'}  Default, 'css3'
        Which set of core tables are being used.  This is used to associate
        column names to known Pisces column definitions, so that text file I/O
        of tables can be supported for a loaded table.
    prefix : str, optional
        All table names will have the provided prefix.
        e.g. "TA_" for "TA_origin", "TA_wfdisc", etc.
    owner : str, optional
        e.g. "myuser" for "myuser.origin", "myuser.wfdisc", etc.  If omitted, 
        the database default owner/account will be used.
    primary_keys : dict, optional
        A mapping of table names to their primary keys of the form,
        {'tablename1': ['primary_key1', 'primary_key2'], ...}
        The SQLAlchemy ORM can only reflect/load tables if they have a primary
        key. If a table is a view or has no primary key, these must be
        specified explicitly.

    Returns
    -------
    dict
        Corresponding dict of mapped SQLAlchemy table classes, keyed by
        canonical name.

    Examples
    --------
    # load standard tables in 'myaccount' table space
    >>> tables = load_tables(session, 'site', 'sitechan', 'wfisc',
    ...                      schema='kbcore', owner='myaccount')

    # load a standard table that has no primary key set
    >>> tables = load_tables(session, 'site', 
    ...                      primary_keys={'site', ['sta', 'ondate']})

    # load all standard css3 tables in the default account
    >>> tables = load_tables(session, schema='css3')


    # load a table that is outside of the schema
    # the table needs to have a primary key defined
    # this will try to use column info definitions known to Pisces
    >>> tables = load_tables(session, 'ccwfdisc', owner='myaccount')

    """
    # TODO: Use sqlalchemy.ext.automap.automap_base inside this function?
    # TODO: After conversion to Python 3, use something better than **kwargs.
    schema = kwargs.get('schema', 'css3')
    prefix = kwargs.get('prefix', '')
    owner = kwargs.get('owner', None)

    # I don't use 'owner' here, b/c it will be enforced by the MetaData. Here,
    # we nust need table names.
    tablenames = make_table_names(*tables, schema=schema, prefix=prefix)

    if owner:
        parents = (declarative_base(metadata=sa.MetaData(schema=owner),
                                    constructor=None, metaclass=PiscesMeta),)
    else:
        parents = ()

    colinfo = getattr(parents, '_column_info_registry', {})

    metadata = sa.MetaData()

    table_classes = dict()
    for table, tablename in tablenames.items():
        itable = sa.Table(tablename, metadata, autoload=True,
                          autoload_with=session.bind, schema=owner)
        if colinfo:
            for col in itable.columns:
                col.info.update(colinfo.get(col.name, {}))

        classdict = {'__table__': itable}

        if primary_keys and tablename in primary_keys:
            columns = [getattr(itable.columns, key) for key in primary_keys[tablename]]
            classdict['__mapper_args__'] = {'primary_key': columns}

        ORMTable = type(tablename.capitalize(), parents, classdict)

        table_classes[table] = ORMTable

    return table_classes


def init_tables(session, *tables):
    """
    Create table classes in a database.

    Parameters
    ----------
    session : sqlalchemy.orm.Session instance
    tables : mapped table classes

    """
    for itable in tables:
        itable.__table__.create(session.bind)
    session.commit()


def drop_tables(session, *tables):
    """
    Drop table classes in a database.

    Parameters
    ----------
    session : sqlalchemy.orm.Session instance
    tables : mapped table classes

    """
    for itable in tables:
        itable.__table__.drop(session.bind)
    session.commit()




from pisces.schema.util import CoreTable
import pisces.schema.css3 as css3


class Affiliation(css3.Affiliation):
    __tablename__ = 'affiliation'

class Amplitude(css3.Amplitude):
    __tablename__ = 'amplitude'

class Arrival(css3.Arrival):
    __tablename__ = 'arrival'

class Assoc(css3.Assoc):
    __tablename__ = 'assoc'

class Event(css3.Event):
    __tablename__ = 'event'

class Gregion(css3.Gregion):
    __tablename__ = 'gregion'

class Instrument(css3.Instrument):
    __tablename__ = 'instrument'

class Lastid(css3.Lastid):
    __tablename__ = 'lastid'

class Netmag(css3.Netmag):
    __tablename__ = 'netmag'

class Network(css3.Network):
    __tablename__ = 'network'

class Origerr(css3.Origerr):
    __tablename__ = 'origerr'

class Origin(css3.Origin):
    __tablename__ = 'origin'

class Remark(css3.Remark):
    __tablename__ = 'remark'

class Sensor(css3.Sensor):
    __tablename__ = 'sensor'

class Site(css3.Site):
    __tablename__ = 'site'

class Sitechan(css3.Sitechan):
    __tablename__ = 'sitechan'

class Sregion(css3.Sregion):
    __tablename__ = 'sregion'

class Stamag(css3.Stamag):
    __tablename__ = 'stamag'
    
class Stassoc(css3.Stassoc):
    __tablename__ = 'stassoc'

class Wfdisc(css3.Wfdisc):
    __tablename__ = 'wfdisc'

class Wftag(css3.Wftag):
    __tablename__ = 'wftag'


CORETABLES = {'affiliation': CoreTable('affiliation', css3.Affiliation, Affiliation),
              'arrival': CoreTable('arrival', css3.Arrival, Arrival),
              'assoc': CoreTable('assoc', css3.Assoc, Assoc),
              'event': CoreTable('event', css3.Event, Event),
              'instrument': CoreTable('instrument', css3.Instrument, Instrument),
              'lastid': CoreTable('lastid', css3.Lastid, Lastid),
              'origin': CoreTable('origin', css3.Origin, Origin),
              'site': CoreTable('site', css3.Site, Site),
              'sitechan': CoreTable('sitechan', css3.Sitechan, Sitechan),
              'wfdisc': CoreTable('wfdisc', css3.Wfdisc, Wfdisc)}



"""
request.py

Convenience functions for common queries.

"""
import numpy as np
from sqlalchemy import func, or_
from obspy.core import UTCDateTime, Stream
import obspy.geodetics as geod

from pisces.io.trace import wfdisc2trace


def get_wfdisc_rows(session, wfdisc, sta=None, chan=None, t1=None, t2=None,
                    wfids=None, daylong=False, asquery=False, verbose=False):
    """
    Returns a list of wfdisc records from provided SQLAlchemy ORM mapped
    wfdisc table, for given station, channel, and time window combination.

    Parameters
    ----------
    session: bound session instance
    wfdisc: SQLAlchemy mapped wfdisc table
    sta, chan, : str, optional
        station, channel strings,
    t1, t2 : int, optional
        Epoch time window of interest (seconds)
        Actually searches for wfdisc.time between t1-86400 and t2 and
        wfdisc.endtime > t1
    wfids : list of integers, optional
        wfid integers. Obviates other arguments.
    daylong : bool, optional
        If True, uses a slightly different time query for best results.
        Not yet implemented (is currently the default behavior).
    asquery : bool, optional
        Return the query object instead of the results.  Default, False.
        Useful if additional you desire additional sorting of filtering.
    verbose : bool, optional
        Print request to the stdout. Not used with asquery=True.

    Returns
    -------
    list of wfdisc row objects, or sqlalchemy.orm.Query instance

    """
    # seconds in a wfdisc file
    CHUNKSIZE = 24 * 60 * 60
    q = session.query(wfdisc)
    if wfids is not None:
        q = q.filter(wfdisc.wfid.in_(wfids))
    else:
        if sta is not None:
            q = q.filter(wfdisc.sta == sta)
        if chan is not None:
            q = q.filter(wfdisc.chan == chan)
        if [t1, t2].count(None) == 0:
            q = q.filter(wfdisc.time.between(t1 - CHUNKSIZE, t2))
            q = q.filter(wfdisc.endtime > t1)
        else:
            if t1 is not None:
                q = q.filter(wfdisc.time >= t1 - CHUNKSIZE)
                q = q.filter(wfdisc.endtime > t1)
            if t2 is not None:
                q = q.filter(wfdisc.time <= t2)

    if asquery:
        res = q
    else:
        if verbose:
            msg = "Requesting sta={}, chan={}, time=[{}, {}], wfids={}"
            print(msg.format(sta, chan, UTCDateTime(t1), UTCDateTime(t2), wfids))
        res = q.all()

    return res


def distaz_query(records, deg=None, km=None, swath=None):
    """
    Out-of-database subset based on distances and/or azimuths.

    Parameters
    ----------
    records : iterable of objects with lat, lon attribute floats
        Target of the subset.
    deg : list or tuple of numbers, optional
        (centerlat, centerlon, minr, maxr)
        minr, maxr in degrees or None for unconstrained.
    km : list or tuple of numbers, optional
        (centerlat, centerlon, minr, maxr)
        minr, maxr in km or None for unconstrained.
    swath : list or tuple of numbers, optional
        (lat, lon, azimuth, tolerance)
        Azimuth (from North) +/-tolerance from lat,lon point in degrees.

    Returns
    -------
    list
        Subset of supplied records.

    """
    #initial True array to propagate through multiple logical AND comparisons
    mask0 = np.ones(len(records), dtype=np.bool)

    if deg:
        dgen = (geod.locations2degrees(irec.lat, irec.lon, deg[0], deg[1]) \
                for irec in records)
        degrees = np.fromiter(dgen, dtype=float)
        if deg[2] is not None:
            mask0 = np.logical_and(mask0, deg[2] <= degrees)
        if deg[3] is not None:
            mask0 = np.logical_and(mask0, deg[3] >= degrees)

        #mask0 = np.logical_and(mask0, mask)

    if km:
        #???: this may be backwards
        mgen = (geod.gps2DistAzimuth(irec.lat, irec.lon, km[0], km[1])[0] \
                  for irec in records)
        kilometers = np.fromiter(mgen, dtype=float)/1e3
        #meters, azs, bazs = zip(*valgen)
        #kilometers = np.array(meters)/1e3
        if km[2] is not None:
            mask0 = np.logical_and(mask0, km[2] <= kilometers)
        if km[3] is not None:
            mask0 = np.logical_and(mask0, km[3] >= kilometers)

        #mask0 = np.logical_and(mask0, mask)

    if swath is not None:
        minaz = swath[2] - swath[3]
        maxaz = swath[2] + swath[3]
        #???: this may be backwards
        azgen = (geod.gps2DistAzimuth(irec.lat, irec.lon, km[0], km[1])[1] \
                 for irec in records)
        azimuths = np.fromiter(azgen, dtype=float)
        mask0 = np.logical_and(mask0, azimuths >= minaz)
        mask0 = np.logical_and(mask0, azimuths <= maxaz)

    #idx = np.nonzero(np.atleast_1d(mask0))[0] ???
    idx = np.nonzero(mask0)[0]
    recs = [records[i] for i in idx]

    return recs


def geographic_query(q, table, region=None, depth=None, asquery=False):
    """
    Filter by region (W, E, S, N) [deg] and/or depth range (min, max) [km].

    """
    if region:
        #(W,E,S,N)
        if region.count(None) == 0:
            q = q.filter(table.lon.between(region[0], region[1]))
            q = q.filter(table.lat.between(region[2], region[3]))
        else:
            if region[0] is not None:
                q = q.filter(table.lon > region[0])
            if region[1] is not None:
                q = q.filter(table.lon < region[1])
            if region[2] is not None:
                q = q.filter(table.lat > region[2])
            if region[3] is not None:
                q = q.filter(table.lat < region[3])

    if depth:
        #(mindep,maxdep)
        if depth.count(None) == 0:
            q = q.filter(table.depth.between(depth[0], depth[1]))
        else:
            if depth[0]:
                q = q.filter(table.depth >= depth[0])
            if depth[1]:
                q = q.filter(table.depth <= depth[1])

    if asquery:
        res = q
    else:
        res = q.all()

    return res

def netstachan_query(q, stations=None, channels=None, networks=None, 
        asquery=False):
    pass

def time_query(q, time=None, daylong=False, asquery=False):
    pass

def get_events(session, origin, event=None, region=None, deg=None, km=None, 
        swath=None, mag=None, depth=None, etime=None, orids=None, evids=None, 
        prefor=False, asquery=False):
    """
    Build common queries for events.

    Parameters
    ----------
    session : sqlalchemy.orm.Session instance
        Must be bound.
    origin : mapped Origin table class
    event : mapped Event table class, optional
    region : list or tuple of numbers, optional
        (W, E, S, N) in degrees. Default, None.
    deg : list or tuple of numbers, optional
        (centerlat, centerlon, minr, maxr) . Default, None.
        minr, maxr in degrees or None for unconstrained.
    km : list or tuple of numbers, optional
        (centerlat, centerlon, minr, maxr)  Default, None.
        minr, maxr in km or None for unconstrained.
    swath : list or tuple of numbers, optional
        (lat, lon, azimuth, tolerance)
        Azimuth (from North) +/-tolerance from lat,lon point in degrees.
        Not yet implemented.
    mag : dict, optional
        {'type1': [min1, max1], 'type2': [min2, max2], ...}
        'type' can be 'mb', 'ms', or 'ml'.  Produces OR clauses.
    depth : tuple or list, optional
        Depth interval [mindep, maxdep] in km.
        Use None for an unconstrained limit.
    etime : tuple or list, optional
        (tstart, tend) epoch event time window
        Use None for an unconstrained limit.
    orids, evids : list or tuple of int, optional
        orid, evid numbers < 1000 in length
        Evids requires event table.
    prefor : bool, optional
        Return preferred origins only. Default False.  Requires event table
        be provided.
    asquery : bool, optional
        Return the query object instead of the results.  Default, False.
        Useful if additional you desire additional sorting of filtering, or
        if you have your own in-database geographic query function(s).  If 
        supplied, deg, km, and/or swath are ignored in the returned query.

    Returns
    -------
    sqlalchemy.orm.Query instance

    Notes
    -----
    Each keyword argument corresponds to an AND clause, except 'mag' which
    returns OR clauses.  Don't submit a request containing both 'evids' and
    'orids' unless you want them joined by an AND clause.  Otherwise process
    them individually, then collate and unique them afterwards.

    """
    Origin = origin
    Event = event

    t = etime

    q = session.query(Origin)

    if orids:
        #[orid1,orid2,...]
        q = q.filter(Origin.orid.in_(orids))

    if t:
        if t.count(None) == 0:
            q = q.filter(Origin.time.between(t[0], t[1]))
        else:
            if t[0]:
                q = q.filter(Origin.time > t[0])
            if t[1]:
                q = q.filter(Origin.time < t[1])

    if mag:
        #TODO: use list comprehension
        magclause = []
        for magtype, vals in mag.iteritems():
            magclause.append(getattr(Origin, magtype).between(vals[0], vals[1]))
        q = q.filter(or_(*magclause))

    if evids:
        q = q.filter(Origin.evid == Event.evid)
        q = q.filter(Event.evid.in_(evids))

    if prefor:
        #q = q.join(Event,Origin.evid==Event.evid)
        #q = q.add_columns(Origin.evid)
        q = q.filter(Origin.orid==Event.prefor)

    q = geographic_query(q, Origin, region=region, depth=depth, asquery=True)

    if asquery:
        res = q
    else:
        res = distaz_query(q.all(), deg=deg, km=km, swath=swath)

    return res
 

def get_stations(session, site, sitechan=None, affiliation=None, stations=None, 
        channels=None, nets=None, loc=None, region=None, deg=None, km=None, 
        swath=None, stime=None, asquery=False):
    """
    Build common queries for stations.

    Parameters
    ----------
    session : sqlalchemy.orm.Session instance
        Must be bound.
    site : mapped Site table class
    sitechan : mapped Sitechan table class, optional
    affiliation : mapped Affiliation table class, optional
    stations : list or tuple of strings
        Desired station code strings.
    channels, nets : list or tuple of strings, or single regex string, optional
        Desired channel, network code strings or regex
    loc : list/tuple, optional
        Location code.
        Not yet implemented.
    region : tuple or list of numbers, optional
        Geographic (W,E,S,N) in degrees, None values for unconstrained.
    deg : list or tuple of numbers, optional
        (centerlat, centerlon, minr, maxr)
        minr, maxr in degrees or None for unconstrained.
    km : list or tuple of numbers, optional
        (centerlat, centerlon, minr, maxr)
        minr, maxr in km or None for unconstrained.
    swath : list or tuple of numbers, optional
        (lat, lon, azimuth, tolerance)
        Azimuth (from North) +/-tolerance from lat,lon point in degrees.
        Currently only works in gnem Oracle.
    asquery : bool, optional
        Return the query object instead of the results.  Default, False.
        Useful if additional you desire additional sorting of filtering, or
        if you have your own in-database geographic query function(s).  If 
        supplied, deg, km, and/or swath are ignored in the returned query.

    Notes
    -----
    Each parameter produces an AND clause, list parameters produce IN 
    clauses, a regex produces a REGEXP_LIKE clause (Oracle-specific?).

    deg, km, and swath are evaluated out-of-database by evaluating all other 
    flags first, then masking.  This can be memory-intensive.  See "Examples"
    for how to perform in-database distance filters.
    
    To include channels or networks with your results use asquery=True, and

    >>> q = q.add_columns(Sitechan.chan)
    >>> q = q.add_columns(Affiliation.net)

    with the returned query.

    Examples
    --------
    Use your own in-database distance query function "km_from_point":

    >>> from sqlalchemy import func
    >>> q = get_stations(session, site, channels=['BHZ'], region=(65,75,30,40), asquery=True)
    >>> stations = q.filter(func.km_from_point(site.lat, site.lon, 40, -110) < 100).all()

    """
    #XXX: remove dependency on func.regexp_like
    Site = site
    Sitechan = sitechan
    Affiliation = affiliation

    d = deg
    t = stime

    q = session.query(Site)

    if stations:
        q = q.filter(Site.sta.in_(sta))

    if nets:
        q = q.join(Affiliation, Affiliation.sta==Site.sta)
        if isinstance(nets, list):
            q = q.filter(Affiliation.net.in_(nets))
        else:
            q = q.filter(func.regexp_like(Affiliation.net, nets))

    if channels:
        q = q.join(Sitechan, Sitechan.sta==Site.sta)
        if isinstance(chans, str):
            #interpret string as regexp
            q = q.filter(func.regexp_like(Sitechan.chan, channnels))
        else:
            q = q.filter(Sitechan.chan.in_(channels))
    
    q = geographic_query(q, Site, region=region, asquery=True)

    if asquery:
        res = q
    else:
        res = distaz_query(q.all(), deg=deg, km=km, swath=swath)

    return res


def get_arrivals(session, arrival, assoc=None, stations=None, channels=None, 
        atime=None, phases=None, arids=None, orids=None, auth=None, 
        asquery=False):
    """
    Build common queries for arrivals.
    
    Parameters
    ----------
    stations, channels : list or tuple of strings
        Desired station, channel strings.
    arrival: mapped Arrival table class
    assoc: mapped Assoc table class, optional
    atime : tuple or list of float, optional
        (tstart, tend) epoch arrival time window. Either can be None.
    phases: list or tuple of strings
        Arrival 'iphase'.
    arids : list of integers
        Desired arid numbers.
    orids : list of integers
        orids from which associated arrivals will be returned. Requires Assoc
        table.
    auth : list/tuple of strings
        Arrival author list.

    Returns
    -------
    list or sqlalchemy.orm.Query instance
        Arrival results.

    Notes
    -----
    Each argument adds an AND clause to the SQL query.
    Unspecified (keyword) arguments are treated as wildcards.  That is, no
    arguments means, "give me all arrivals everywhere ever."

    """
    Arrival = arrival
    Assoc = assoc

    t = atime

    q = session.query(Arrival)

    if stations:
        q = q.filter(Arrival.sta.in_(stations))

    if channels:
        q = q.filter(Arrival.chan.in_(channels))

    if phases:
        q = q.filter(Arrival.iphase.in_(phase))

    if t:
        if t.count(None) == 0:
            q = q.filter(Arrival.time.between(t[0], t[1]))
        else:
            if t[0]:
                q = q.filter(Arrival.time > t[0])
            if t[1]:
                q = q.filter(Arrival.time < t[1])

    if arids:
        q = q.filter(Arrival.arid.in_(arids))

    if orids:
        q = q.filter(Arrival.arid == Assoc.arid)
        q = q.filter(Assoc.orid.in_(orids))

    if auth:
        q = q.filter(Arrival.auth.in_(auth))

    if asquery:
        res = q
    else:
        res = q.all()

    return res


def get_waveforms(session, wfdisc, station=None, channel=None, starttime=None,
                  endtime=None, wfids=None, tol=None):
    """
    Request waveforms.

    Parameters
    ----------
    session : sqlalchemy.orm.Session instance
        Must be bound.
    wfdisc : mapped Wfdisc table class
    station, channel : str, optional
        Desired station, channel code strings
    starttimes, endtimes : float, optional
        Epoch start times, end times.
        Traces will be cut to these times.
    wfids : iterable of int, optional
        Wfdisc wfids.  Obviates the above arguments and just returns full Wfdisc
        row waveforms.
    tol : float
        If provided, a warning is fired if any Trace is not within tol seconds
        of starttime and endtime.

    Returns
    -------
    obspy.Stream
        Traces are merged and cut to requested times.

    Raises
    ------
    ValueError
        Returned Stream contains trace start/end times outside of the tolerance.

    """
    # TODO: add evids= option?, use with stawin= option in .execute method?
    # TODO: implement get_arrivals if arrivals=True
    Wfdisc = wfdisc

    st = Stream()
    if wfids:
        station = channel = starttime = endtime = None

    starttime = float(starttime) if starttime is not None else None
    endtime = float(endtime) if endtime is not None else None

    t1_utc = UTCDateTime(starttime) if starttime is not None else None
    t2_utc = UTCDateTime(endtime) if endtime is not None else None

    wfs = get_wfdisc_rows(session, Wfdisc, station, channel, starttime, endtime,
                          wfids=wfids)
    # TODO:
    # Maybe a warning can be fired if the get_waveforms(session, Wfdisc,
    # starttime=t_S, endtime=t_E) function gets data that does not cover
    # starttime endtime (within a tolerance)
    for wf in wfs:
        try:
            tr = wfdisc2trace(wf)
        except IOError:
            # can't read file
            tr = None

        if tr:
            # None utc times will pass through
            tr.trim(t1_utc, t2_utc)
            st.append(tr)
            # TODO: do arrival stuff here?

    if all([tol, starttime, endtime]):
        starttimes, endtimes = zip(*[(t.stats.starttime, t.stats.endtime) for t in st])
        min_t = float(min(starttimes))
        max_t = float(max(endtimes))
        if (abs(min_t - starttime) > tol) or (abs(max_t - endtime) > tol):
            msg = "Trace times are outside of tolerance: {}".format(tol)
            # XXX: change this to a real Pisces exception
            raise ValueError(msg)


    return st


def get_ids(session, lastid, ids, detach=False):
    """
    Get or create lastid rows.

    Parameters
    ----------
    session : sqlalchemy.orm.session instance, bound
    lastid : sqlalchemy orm mapped lastid table
    ids : list
        Desired lastid keyname strings.
    detach : bool, optional
        If True, expunge results from session before returning.
        Useful if you don't have permission on lastid, and don't want
        session commits to throw a permission error.


    Returns
    -------
    list
        Corresponding existing or new rows from lastid table.

    Notes
    -----
    Keyvalue is 0 if id name not present in lastid table.

    """
    out = []
    for idname in ids:
        iid = session.query(lastid).filter(lastid.keyname == idname).first()
        if not iid:
            iid = lastid(keyname=idname, keyvalue=0)
        out.append(iid.keyvalue)

    if detach:
        session.expunge_all(out)

    return out



"""
Conversions between Trace objects and table-like objects.

"""
import os
from obspy.core import UTCDateTime, Trace, Stats
from pisces.io.readwaveform import read_waveform
from pisces.io.util import _buildhdr


def wfdisc2obspyhdr(wf):
    """ ObsPy Stats dict from Wfdisc instance.
    """
    keymap = {'npts': 'nsamp', 'calib': 'calib', 'channel': 'chan',
              'sampling_rate': 'samprate', 'station': 'sta'}
    obshdr = _buildhdr(keymap, wf)
    obshdr['starttime'] = UTCDateTime(float(wf.time))
    obshdr['delta'] = 1. / wf.samprate

    return Stats(header=obshdr)


def wfdisc2trace(wf):
    """
    Takes a fielded wfdisc row record and produces a minimal obspy Trace with
    some SAC header fields included, or None.

    Raises
    ------
    IOError
        Can't read the waveform file.

    """
    hdr = wfdisc2obspyhdr(wf)

    data = read_waveform(os.sep.join([wf.dir, wf.dfile]), wf.datatype, wf.foff,
                         wf.nsamp)

    if data is not None:
        tr = Trace(data=data, header=hdr)
    else:
        tr = None

    return tr



# Converted to Python 3.5.2 on 01-25-17
# by: Jeremy Webster

# coding: utf-8
"""
Center for Seismic Studies relational database schema 3.0 (CSS3.0)


"""
from datetime import datetime
from sqlalchemy import DateTime, Float, String, Integer
from sqlalchemy import Column
from sqlalchemy import PrimaryKeyConstraint, UniqueConstraint
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.declarative import declared_attr

from obspy.core import UTCDateTime
from pisces.schema.util import PiscesMeta
from pisces.schema.util import parse_int, parse_float, parse_str

from pisces.io.trace import wfdisc2trace

Base = declarative_base(metaclass=PiscesMeta, constructor=None)

# COLUMN DEFINITIONS
# Generic SQLA types, compatible with different backends
# !! info dictionary defines the external representations (NumPy, text files)
#   and default values for the mapped class representation.
#
# XXX: for numeric types, maximum width is not enforced!
#
# NOTE: info['dtype'] for floats/ints should match the system default for Python
#       "{:f}".format(numpyfloat/int) to work, b/c it can use the builtin float
#       __format__().  use 'float' or 'int'
#       See:
#       http://stackoverflow.com/questions/16928644/floats-in-numpy-structured-array-and-native-string-formatting-with-format


def strip(s):
    return str(s).strip()


DATEFMT = '%y-%m-%d %H:%M:%S'


def dtfn(s):
    try:
        dt = datetime.strptime(s.strip(), DATEFMT)
    except ValueError:
        # Antelope convention
        dt = UTCDateTime(float(s)).datetime
    return dt


algorithm = Column(String(15),
                   info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15.15s'})

amp = Column(Float(24),
             info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 10, 'format': '10.1f'})

ampid = Column(Integer, nullable=False,
               info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 9, 'format': '9d'})

amptime = Column(Float(53),
                 info={'default': -9999999999.999, 'parse': parse_float, 'dtype': 'float', 'width': 17, 'format': '17.5f'})

amptype = Column(String(8),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})

arid = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

auth = Column(String(15),
              info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15s'})

azdef = Column(String(1),
               info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

azimuth = Column(Float(24),
                 info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

azres = Column(Float(24),
               info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.1f'})

band = Column(String(1),
              info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

belief = Column(Float(24),
                info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 4, 'format': '4.2f'})

calib = Column(Float(24),
               info={'default': 1.0, 'parse': parse_float, 'dtype': 'float', 'width': 16, 'format': '16.6f'})

calper = Column(Float(24),
                info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 16, 'format': '16.6f'})

calratio = Column(Float(24),
                  info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 16, 'format': '16.6f'})

chan = Column(String(8),
              info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})

chanid = Column(Integer,
                info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

clip = Column(String(1),
              info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

commid = Column(Integer,
                info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

conf = Column(Float(24),
              info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 5, 'format': '5.3f'})

ctype = Column(String(4),
               info={'default': '-', 'parse': parse_str, 'dtype': 'a4', 'width': 4, 'format': '4.4s'})

datatype = Column(String(2),
                  info={'default': '-', 'parse': parse_str, 'dtype': 'a2', 'width': 2, 'format': '2.2s'})

deast = Column(Float(24),
               info={'default': 0.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

delaz = Column(Float(24),
               info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

delslo = Column(Float(24),
                info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

delta = Column(Float(24),
               info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 8, 'format': '8.3f'})

deltaf = Column(Float(24),
                info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.3f'})

deltim = Column(Float(24),
                info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.3f'})

depdp = Column(Float(24),
               info={'default': -999, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

depth = Column(Float(24),
               info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

descrip = Column(String(50),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a50', 'width': 50, 'format': '50.50s'})

dfile = Column(String(32),
               info={'default': '-', 'parse': parse_str, 'dtype': 'a32', 'width': 32, 'format': '32.32s'})

digital = Column(String(1),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

dir = Column(String(64),
             info={'default': '-', 'parse': parse_str, 'dtype': 'a64', 'width': 64, 'format': '64.64s'})

dist = Column(Float(24),
              info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

dnorth = Column(Float(24),
                info={'default': 0.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

dtype = Column(String(1),
               info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

duration = Column(Float(24),
                  info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

edepth = Column(Float(24),
                info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

elev = Column(Float(24),
              info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

ema = Column(Float(24),
             info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

emares = Column(Float(24),
                info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.1f'})

endtime = Column(Float(53),
                 info={'default': 9999999999.999, 'parse': parse_float, 'dtype': 'float', 'width': 17, 'format': '17.5f'})

esaz = Column(Float(24),
              info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

etype = Column(String(7),
               info={'default': '-', 'parse': parse_str, 'dtype': 'a7', 'width': 7, 'format': '7.7s'})

evid = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

evname = Column(String(15),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15.15s'})

fm = Column(String(2),
            info={'default': '-', 'parse': parse_str, 'dtype': 'a2', 'width': 2, 'format': '2.2s'})

foff = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 10, 'format': '10d'})

grn = Column(Integer,
             info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

grname = Column(String(40),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a40', 'width': 40, 'format': '40.40s'})

hang = Column(Float(24),
              info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.1f'})

imb = Column(Float(24),
             info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

iml = Column(Float(24),
             info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

ims = Column(Float(24),
             info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

inarrival = Column(String(1),
                   info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

inid = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

insname = Column(String(50),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a50', 'width': 50, 'format': '50.50s'})

instant = Column(String(1),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

instype = Column(String(6),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a6', 'width': 6, 'format': '6.6s'})

iphase = Column(String(8),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})

jdate = Column(Integer,
               info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

keyname = Column(String(15),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15.15s'})

keyvalue = Column(Integer,
                  info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

lat = Column(Float(24),
             info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

lddate = Column(DateTime, nullable=False, onupdate=datetime.now,
                info={'default': datetime.now, 'parse': dtfn, 'dtype': 'O', 'width': 17, 'format': DATEFMT})

lineno = Column(Integer,
                info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

location = Column(String(32),
                  info={'default': '-', 'parse': parse_str, 'dtype': 'a32', 'width': 32, 'format': '32.32s'})

logat = Column(Float(24),
               info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

lon = Column(Float(24),
             info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

magid = Column(Integer, nullable=False,
               info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

magnitude = Column(Float(24),
                   info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

magres = Column(Float(24),
                info={'default': -999, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

magdef = Column(String(1),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

magtype = Column(String(6),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a6', 'width': 6, 'format': '6.6s'})

mb = Column(Float(24),
            info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

mbid = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

ml = Column(Float(24),
            info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

mlid = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

ms = Column(Float(24),
            info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

msid = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

mmodel = Column(String(15),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15.15s'})

nass = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 4, 'format': '4d'})

ncalib = Column(Float(24),
                info={'default': 1.0, 'parse': parse_float, 'dtype': 'float', 'width': 16, 'format': '16.6f'})

ncalper = Column(Float(24),
                 info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 16, 'format': '16.6f'})

ndef = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 4, 'format': '4d'})

ndp = Column(Integer,
             info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 4, 'format': '4d'})

net = Column(String(8),
             info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})

netname = Column(String(80),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a80', 'width': 80, 'format': '80.80s'})

nettype = Column(String(4),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a4', 'width': 4, 'format': '4.4s'})

nsamp = Column(Integer,
               info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

nsta = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

offdate = Column(Integer,
                 info={'default': 2286324, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

ondate = Column(Integer,
                info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

orid = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

parid = Column(Integer,
               info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 9, 'format': '9d'})

per = Column(Float(24),
             info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

phase = Column(String(8),
               info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})

prefor = Column(Integer,
                info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

qual = Column(String(1),
              info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

rect = Column(Float(24),
              info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.3f'})

refsta = Column(String(6),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a6', 'width': 6, 'format': '6.6s'})

remark = Column(String(80),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a80', 'width': 80, 'format': '80.80s'})

rsptype = Column(String(6),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a6', 'width': 6, 'format': '6.6s'})

samprate = Column(Float(24),
                  info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 11, 'format': '11.7f'})

sdepth = Column(Float(24),
                info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

sdobs = Column(Float(24),
               info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

seaz = Column(Float(24),
              info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

segtype = Column(String(1),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

slodef = Column(String(1),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

slores = Column(Float(24),
                info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

slow = Column(Float(24),
              info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

smajax = Column(Float(24),
                info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

sminax = Column(Float(24),
                info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

snr = Column(Float(24),
             info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 10, 'format': '10.2f'})

srn = Column(Integer,
             info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

srname = Column(String(40),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a40', 'width': 40, 'format': '40.40s'})

sta = Column(String(6),
             info={'default': '-', 'parse': parse_str, 'dtype': 'a6', 'width': 6, 'format': '6.6s'})

staname = Column(String(50),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a50', 'width': 50, 'format': '50.50s'})

stassid = Column(Integer,
                 info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

statype = Column(String(4),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a4', 'width': 4, 'format': '4.4s'})

stime = Column(Float(24),
               info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 8, 'format': '8.2f'})

strike = Column(Float(24),
                info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.2f'})

stt = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

stx = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

sty = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

stype = Column(String(1),
               info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

stz = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

sxx = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

sxy = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

sxz = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

syy = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

syz = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

szz = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

tagid = Column(Integer,
               info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 8, 'format': '8d'})

tagname = Column(String(8),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})

time = Column(Float(53),
              info={'default': -9999999999.999, 'parse': parse_float, 'dtype': 'float', 'width': 17, 'format': '17.5f'})

timedef = Column(String(1), info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

timeres = Column(Float(24),
                 info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 8, 'format': '8.3f'})

tshift = Column(Float(24),
                info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.2f'})

uncertainty = Column(Float(24),
                     info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

units = Column(String(15),
               info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15.15s'})

vang = Column(Float(24),
              info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.1f'})

vmodel = Column(String(15),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15.15s'})

wfid = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

wgt = Column(Float(24),
             info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.3f'})


class Affiliation(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('net', 'sta'),)

    net = net.copy()
    sta = sta.copy()
    lddate = lddate.copy()


class Amplitude(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('ampid'),)

    ampid = ampid.copy()
    arid = arid.copy()
    parid = parid.copy()
    chan = chan.copy()
    amp = amp.copy()
    per = per.copy()
    snr = snr.copy()
    amptime = amptime.copy()
    time = time.copy()
    duration = duration.copy()
    deltaf = deltaf.copy()
    amptype = amptype.copy()
    units = units.copy()
    clip = clip.copy()
    inarrival = inarrival.copy()
    auth = auth.copy()
    lddate = lddate.copy()


class Arrival(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('arid'),
                UniqueConstraint('sta', 'time', 'chan', 'iphase', 'auth'),)

    sta = sta.copy()
    time = time.copy()
    arid = arid.copy()
    jdate = jdate.copy()
    stassid = stassid.copy()
    chanid = chanid.copy()
    chan = chan.copy()
    iphase = iphase.copy()
    stype = stype.copy()
    deltim = deltim.copy()
    azimuth = azimuth.copy()
    delaz = delaz.copy()
    slow = slow.copy()
    delslo = delslo.copy()
    ema = ema.copy()
    rect = rect.copy()
    amp = amp.copy()
    per = per.copy()
    logat = logat.copy()
    clip = clip.copy()
    fm = fm.copy()
    snr = snr.copy()
    qual = qual.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Assoc(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('arid', 'orid'), UniqueConstraint('arid'), )

    arid = arid.copy()
    orid = orid.copy()
    sta = sta.copy()
    phase = phase.copy()
    belief = belief.copy()
    delta = delta.copy()
    seaz = seaz.copy()
    esaz = esaz.copy()
    timeres = timeres.copy()
    timedef = timedef.copy()
    azres = azres.copy()
    azdef = azdef.copy()
    slores = slores.copy()
    slodef = slodef.copy()
    emares = emares.copy()
    wgt = wgt.copy()
    vmodel = vmodel.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Event(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('evid'), UniqueConstraint('prefor'),)

    evid = evid.copy()
    evname = evname.copy()
    prefor = prefor.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Gregion(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('grn'),)

    grn = grn.copy()
    grname = grname.copy()
    lddate = lddate.copy()


class Instrument(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('inid'),)

    inid = inid.copy()
    insname = insname.copy()
    instype = instype.copy()
    band = band.copy()
    digital = digital.copy()
    samprate = samprate.copy()
    ncalib = ncalib.copy()
    ncalper = ncalper.copy()
    dir = dir.copy()
    dfile = dfile.copy()
    rsptype = rsptype.copy()
    lddate = lddate.copy()


class Lastid(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('keyname'), UniqueConstraint('keyname', 'keyvalue'),)

    keyname = keyname.copy()
    keyvalue = keyvalue.copy()
    lddate = lddate.copy()


class Netmag(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('magid'),
                UniqueConstraint('magid', 'orid'),)

    magid = magid.copy()
    net = net.copy()
    orid = orid.copy()
    evid = evid.copy()
    magtype = magtype.copy()
    nsta = nsta.copy()
    magnitude = magnitude.copy()
    uncertainty = uncertainty.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Network(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('net'),)

    net = net.copy()
    netname = netname.copy()
    nettype = nettype.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Origerr(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('orid'),)

    orid = orid.copy()
    sxx = sxx.copy()
    syy = syy.copy()
    szz = szz.copy()
    stt = stt.copy()
    sxy = sxy.copy()
    sxz = sxz.copy()
    syz = syz.copy()
    stx = stx.copy()
    sty = sty.copy()
    stz = stz.copy()
    sdobs = sdobs.copy()
    smajax = smajax.copy()
    sminax = sminax.copy()
    strike = strike.copy()
    sdepth = sdepth.copy()
    stime = stime.copy()
    conf = conf.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Origin(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (UniqueConstraint('lat', 'lon', 'depth', 'time', 'auth'), PrimaryKeyConstraint('orid'))

    lat = lat.copy()
    lon = lon.copy()
    depth = depth.copy()
    time = time.copy()
    orid = orid.copy()
    evid = evid.copy()
    jdate = jdate.copy()
    nass = nass.copy()
    ndef = ndef.copy()
    ndp = ndp.copy()
    grn = grn.copy()
    srn = srn.copy()
    etype = etype.copy()
    depdp = depdp.copy()
    dtype = dtype.copy()
    mb = mb.copy()
    mbid = mbid.copy()
    ms = ms.copy()
    msid = msid.copy()
    ml = ml.copy()
    mlid = mlid.copy()
    algorithm = algorithm.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Remark(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('commid', 'lineno'),)

    commid = commid.copy()
    lineno = lineno.copy()
    remark = remark.copy()
    lddate = lddate.copy()


class Sensor(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('sta', 'chan', 'time', 'endtime'),)

    sta = sta.copy()
    chan = chan.copy()
    time = time.copy()
    endtime = endtime.copy()
    inid = inid.copy()
    chanid = chanid.copy()
    jdate = jdate.copy()
    calratio = calratio.copy()
    calper = calper.copy()
    tshift = tshift.copy()
    instant = instant.copy()
    lddate = lddate.copy()


class Site(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('sta', 'ondate'),)

    sta = sta.copy()
    ondate = ondate.copy()
    offdate = offdate.copy()
    lat = lat.copy()
    lon = lon.copy()
    elev = elev.copy()
    staname = staname.copy()
    statype = statype.copy()
    refsta = refsta.copy()
    dnorth = dnorth.copy()
    deast = deast.copy()
    lddate = lddate.copy()


class Sitechan(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (UniqueConstraint('sta', 'chan', 'ondate'),
                PrimaryKeyConstraint('chanid'),)

    sta = sta.copy()
    chan = chan.copy()
    ondate = ondate.copy()
    chanid = chanid.copy()
    offdate = offdate.copy()
    ctype = ctype.copy()
    edepth = edepth.copy()
    hang = hang.copy()
    vang = vang.copy()
    descrip = descrip.copy()
    lddate = lddate.copy()


class Sregion(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('srn'),)

    srn = srn.copy()
    srname = srname.copy()
    lddate = lddate.copy()


class Stamag(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('magid', 'sta'),)

    magid = magid.copy()
    sta = sta.copy()
    arid = arid.copy()
    orid = orid.copy()
    evid = evid.copy()
    phase = phase.copy()
    delta = delta.copy()
    magtype = magtype.copy()
    magnitude = magnitude.copy()
    uncertainty = uncertainty.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()
    
    
class Stassoc(Base):
    __abstract__ = True
    
    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('stassid'),)
        
    stassid = stassid.copy()
    sta = sta.copy()
    etype = etype.copy()
    location = location.copy()
    dist = dist.copy()
    azimuth = azimuth.copy()
    lat = lat.copy()
    lon = lon.copy()
    depth = depth.copy()
    time = time.copy()
    imb = imb.copy()
    ims = ims.copy()
    iml = iml.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()
        

class Wfdisc(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (UniqueConstraint('wfid', 'dir', 'dfile'), PrimaryKeyConstraint('wfid'),)

    def to_trace(self):
        """
        Read the wfdisc line into a Trace instance.  Minimal header.

        Returns
        -------
        obspy.Trace

        """
        return wfdisc2trace(self)

    sta = sta.copy()
    chan = chan.copy()
    time = time.copy()
    wfid = wfid.copy()
    chanid = chanid.copy()
    jdate = jdate.copy()
    endtime = endtime.copy()
    nsamp = nsamp.copy()
    samprate = samprate.copy()
    calib = calib.copy()
    calper = calper.copy()
    instype = instype.copy()
    segtype = segtype.copy()
    datatype = datatype.copy()
    clip = clip.copy()
    dir = dir.copy()
    dfile = dfile.copy()
    foff = foff.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Wftag(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('tagid'), UniqueConstraint('tagname', 'tagid', 'wfid'),)

    tagname = tagname.copy()
    tagid = tagid.copy()
    wfid = wfid.copy()
    lddate = lddate.copy()



# Converted to python 3.5.2 on 02/07/17
# By: Jeremy Webster

"""
Conversions between SAC header variables and CSS or KB Core fields.


Converts a SAC header dictionary into a list of table dictionaries and vice-versa.

"""
# XXX: currently not working
# TODO:
#   - remove functions already in pisces.io.trace
#   - make everything just translate dictionaries, not classes and make
#     everything less obspy.Trace-centric
#   - change db.flatfile.KBTABLEDICT to use pisces.schema.util.get_infovals
#   - use pysac (hopefully obspy.io.sac, eventually)
#   - make a bunch of column/header translation functions, like "kcmpnm_to_chan"
#     and "chan_to_kcmpnm", and use them in higher-level functions, like
#     "sachdr_to_wfdisc" and "wfdisc_to_sachdr".

import sys
import os
from collections import OrderedDict
import functools

from obspy.core import UTCDateTime, AttribDict
import obspy.geodetics as geod

import pisces.tables.kbcore as kb
from pisces.io.readwaveform import read_waveform
from pisces.io.util import _map_header, _buildhdr

# ObsPy default values
OBSPYDEFAULT = {'network': '',
                'station': '',
                'location': '',
                'channel': ''}

# TODO: much of this will be superseded by obspy.io.sac
# ------------------ CONVERT SAC HEADER DICTIONARY TO TABLES ------------#
# SAC default values
IDEFAULT = -12345
FDEFAULT = -12345.0
SDEFAULT = '-12345  '
SLDEFAULT = '-12345          '
SACDEFAULT = {'a': FDEFAULT, 'az': FDEFAULT, 'b': FDEFAULT, 'baz': FDEFAULT,
              'cmpaz': FDEFAULT, 'cmpinc': FDEFAULT, 'delta': FDEFAULT,
              'depmax': FDEFAULT, 'depmen': FDEFAULT, 'depmin': FDEFAULT,
              'dist': FDEFAULT, 'e': FDEFAULT, 'evdp': FDEFAULT, 'evla': FDEFAULT,
              'evlo': FDEFAULT, 'f': FDEFAULT, 'gcarc': FDEFAULT, 'idep': IDEFAULT,
              'ievreg': IDEFAULT, 'ievtype': IDEFAULT, 'iftype': IDEFAULT,
              'iinst': IDEFAULT, 'imagsrc': IDEFAULT, 'imagtyp': IDEFAULT,
              'int1': FDEFAULT, 'iqual': IDEFAULT, 'istreg': IDEFAULT,
              'isynth': IDEFAULT, 'iztype': IDEFAULT, 'ka': SDEFAULT,
              'kcmpnm': SDEFAULT, 'kdatrd': SDEFAULT, 'kevnm': SLDEFAULT,
              'kf': SDEFAULT, 'khole': SDEFAULT, 'kinst': SDEFAULT,
              'knetwk': SDEFAULT, 'ko': SDEFAULT, 'kstnm': SDEFAULT, 'kt0': SDEFAULT,
              'kt1': SDEFAULT, 'kt2': SDEFAULT, 'kt3': SDEFAULT, 'kt4': SDEFAULT,
              'kt5': SDEFAULT, 'kt6': SDEFAULT, 'kt7': SDEFAULT, 'kt8': SDEFAULT,
              'kt9': SDEFAULT, 'kuser0': SDEFAULT, 'kuser1': SDEFAULT,
              'kuser2': SDEFAULT, 'lcalda': IDEFAULT, 'leven': IDEFAULT,
              'lovrok': IDEFAULT, 'lpspol': IDEFAULT, 'mag': FDEFAULT,
              'nevid': IDEFAULT, 'norid': IDEFAULT, 'npts': IDEFAULT,
              'nvhdr': IDEFAULT, 'nwfid': IDEFAULT, 'nzhour': IDEFAULT,
              'nzjday': IDEFAULT, 'nzmin': IDEFAULT, 'nzmsec': IDEFAULT,
              'nzsec': IDEFAULT, 'nzyear': IDEFAULT, 'o': FDEFAULT,
              'odelta': FDEFAULT, 'scale': FDEFAULT, 'stdp': FDEFAULT,
              'stel': FDEFAULT, 'stla': FDEFAULT, 'stlo': FDEFAULT, 't0': FDEFAULT,
              't1': FDEFAULT, 't2': FDEFAULT, 't3': FDEFAULT, 't4': FDEFAULT,
              't5': FDEFAULT, 't6': FDEFAULT, 't7': FDEFAULT, 't8': FDEFAULT,
              't9': FDEFAULT, 'unused10': FDEFAULT, 'unused11': FDEFAULT,
              'unused12': FDEFAULT, 'unused6': FDEFAULT, 'unused7': FDEFAULT,
              'unused8': FDEFAULT, 'unused9': FDEFAULT, 'user0': FDEFAULT,
              'user1': FDEFAULT, 'user2': FDEFAULT, 'user3': FDEFAULT,
              'user4': FDEFAULT, 'user5': FDEFAULT, 'user6': FDEFAULT,
              'user7': FDEFAULT, 'user8': FDEFAULT, 'user9': FDEFAULT,
              'xmaximum': FDEFAULT, 'xminimum': FDEFAULT, 'ymaximum': FDEFAULT,
              'yminimum': FDEFAULT}

############################# DECORATORS ######################################
# Decorator functions allow readable, reusable handling of header values.


def cast_to_int(original_func):
    """
    Cast a function's argument to int before the function operates on it, like:

    Examples
    --------
    >>> @cast_to_int
    ... def nwfid_to_wfid(nwfid):
    ...     return nwfid
    >>> nwfid_to_wfid(12.3)
    12

    """
    @functools.wraps(original_func)
    def converter(hdr):
        return original_func(int(hdr))
    return converter


def cast_to_float(original_func):
    """
    Cast a function's argument to float before the function operates on it, like:

    Examples
    --------
    >>> @cast_to_float
    ... def stla_to_lat(stla):
    ...     return stla
    >>> lat = stla_to_lat(numpy.float32(12.3))
    >>> type(lat)
    float
    # not numpy.float32

    """
    @functools.wraps(original_func)
    def converter(hdr):
        return original_func(float(hdr))
    return converter


def strip_string(original_func):
    """
    Strip white space from a function's argument to float before the function
    operates on it, like:

    Examples
    --------
    >>> @strip_string
    ... def kcmpnm_to_chan(kcmpnm):
    ...     return kcmpnm[:6]
    >>> kclmpnm_to_chan(' my_too_long_component ')
    'my_too'

    """
    @functools.wraps(original_func)
    def converter(hdr):
        return original_func(int(hdr))
    return converter


def truncate_string(N):
    def make_func(original_func):
        @functools.wraps(original_func)
        def func_wrapper(val):
            return val[:N]
        return func_wrapper


# decorator for when you find default/null SAC header values
# if a value is detected, function returns return_value instead of being executed
# XXX: not clear how this could be used with other decorators
def swap_if_value(detected_value, return_value):
    def check_value(original_func):
        def func_wrapper(arg):
            if arg == detected_value:
                out = return_value
            else:
                out = arg
            return out
        return func_wrapper


# ############################## STRING HEADER CONVERSIONS ####################
# SAC -> CSS
def kcmpnm_to_chan(kcmpnm):
    return kcmpnm.strip()[:8]


def kevnm_to_evname(kevnm):
    return kevnm.strip()


def kinst_to_insname(kinst):
    return kinst.strip()


def knetwk_to_net(knetwk):
    return knetwk.strip()[:8]


def kstnm_to_sta(kstnm):
    return kstnm.strip()[:6]


# CSS -> SAC
def chan_to_kcmpnm(chan):
    return chan


def evname_to_kevnm(evname):
    return evname


def insname_to_kinst(insname):
    return insname


def net_to_knetwk(net):
    return net


def sta_to_kstnm(sta):
    return kstnm


# ############################## INT HEADER CONVERSIONS #######################
# DATA
#
ENUM_NAMES = {1: 'itime', 2: 'irlim', 3: 'iamph', 4: 'ixy', 5: 'iunkn',
              6: 'idisp', 7: 'ivel', 8: 'iacc', 9: 'ib', 10: 'iday', 11: 'io',
              12: 'ia', 13: 'it0', 14: 'it1', 15: 'it2', 16: 'it3', 17: 'it4',
              18: 'it5', 19: 'it6', 20: 'it7', 21: 'it8', 22: 'it9',
              23: 'iradnv', 24: 'itannv', 25: 'iradev', 26: 'itanev',
              27: 'inorth', 28: 'ieast', 29: 'ihorza', 30: 'idown', 31: 'iup',
              32: 'illlbb', 33: 'iwwsn1', 34: 'iwwsn2', 35: 'ihglp', 36: 'isro',
              37: 'inucl', 38: 'ipren', 39: 'ipostn', 40: 'iquake', 41: 'ipreq',
              42: 'ipostq', 43: 'ichem', 44: 'iother', 45: 'igood', 46: 'iglch',
              47: 'idrop', 48: 'ilowsn', 49: 'irldta', 50: 'ivolts', 52: 'imb',
              53: 'ims', 54: 'iml', 55: 'imw', 56: 'imd', 57: 'imx', 58: 'ineic',
              59: 'ipdeq', 60: 'ipdew', 61: 'ipde', 62: 'iisc', 63: 'ireb',
              64: 'iusgs', 65: 'ibrk', 66: 'icaltech', 67: 'illnl', 68: 'ievloc',
              69: 'ijsop', 70: 'iuser', 71: 'iunknown', 72: 'iqb', 73: 'iqb1',
              74: 'iqb2', 75: 'iqbx', 76: 'iqmt', 77: 'ieq', 78: 'ieq1',
              79: 'ieq2', 80: 'ime', 81: 'iex', 82: 'inu', 83: 'inc', 84: 'io_',
              85: 'il', 86: 'ir', 87: 'it', 88: 'iu', 89: 'ieq3', 90: 'ieq0',
              91: 'iex0', 92: 'iqc', 93: 'iqb0', 94: 'igey', 95: 'ilit',
              96: 'imet', 97: 'iodor', 103: 'ios'}

# ievtyp -> etype
ETYPEDICT = {37: 'en', 38: 'ex', 39: 'ex', 40: 'qt', 41: 'qt', 42: 'qt',
             43: 'ec', 72: 'me', 73: 'me', 74: 'me', 75: 'me', 76: 'mb',
             77: 'qt', 78: 'qt', 79: 'qt', 80: 'ex', 81: 'ex', 82: 'en',
             83: 'mc'}

# evtype -> ievtyp
IEVTYPDICT = dict((_val, _key) for _key, _val in ETYPEDICT.items())
#
# imagsrc -> auth
AUTHDICT = {58: 'ISC:NEIC', 61: 'PDE', 62: 'ISC', 63: 'REB-ICD',
            64: 'IUSGS', 65: 'ISC:BERK', 66: 'ICALTECH', 67: 'ILLNL',
            68: 'IEVLOC', 69: 'IJSOP', 70: 'IUSER', 71: 'IUNKNOWN'}
#
# auth -> imagsrc
IMAGSRCDICT = dict((_val, _key) for _key, _val in AUTHDICT.items())


# SAC -> CSS
def ievreg_to_grn(ievreg):
    return int(ievreg)


def ievtyp_to_etype(ievtyp):
    """Provide the ievtyp(e) integer, get the etype string."""
    return ETYPEDICT[ievtyp]


# the following functions accept a SAC header dictionary, and return respective
# kbcore table instances, assumes default SAC header values set to None
def get_sac_reftime(header):
    """
    Get SAC header reference time as a UTCDateTime instance from a SAC header
    dictionary.

    If using ObsPy to read the SAC file, use debug_headers=True to get the full
    header, including nz time headers.

    """
    # "nz" fields are not kept when ObsPy reads a SAC file.  They're used
    # to make "starttime", then discarded.
    # t0 = nzyear + nzjday + nzhour + nzminute + nzsecond + nzmsec*1000
    # tr.stats.starttime = t0 + b
    # therefore: t0 = starttime - b

    # TODO: let null nz values be 0?
    try:
        yr = header['nzyear']
        if 0 <= yr <= 99:
            yr += 1900
        nzjday = header['nzjday']
        nzhour = header['nzhour']
        nzmin = header['nzmin']
        nzsec = header['nzsec']
        nzmsec = header['nzmsec']
    except KeyError as e:
        msg = "Not enough time information: {}".format(e.message)
        raise KeyError(msg)

    try:
        reftime = UTCDateTime(year=yr, julday=nzjday, hour=nzhour, minute=nzmin,
                              second=nzsec, microsecond=nzmsec * 1000)
        # reftime = datetime.datetime(yr, 1, 1, nzhour, nzmin, nzsec, nzmsec * 1000) + \
        #                            datetime.timedelta(int(nzjday-1))
        # NOTE: epoch seconds can be got by:
        # (reftime - datetime.datetime(1970,1,1)).total_seconds()
    except ValueError:
        # may contain -12345 null values?
        msg = "Invalid or missing time headers."
        raise ValueError(msg)

    return reftime


def _cast_int(d, keys):
    for key in keys:
        try:
            d[key] = int(d[key])
        except (TypeError, KeyError):
            # no key, or int(None)
            pass

    return d


def _cast_float(d, keys):
    for key in keys:
        try:
            d[key] = float(d[key])
        except (TypeError, KeyError):
            # no key, or int(None)
            pass

    return d


def _clean_str(d, keys):
    for key in keys:
        try:
            d[key] = d[key].strip()
        except (KeyError, AttributeError):
            # no key, None.strip()
            pass

    return d


def sachdr2site(header):
    """
    Provide a SAC header dictionary, get a site table dictionary.

    """
    sac_site = [('kstnm', 'sta'),
                ('stla', 'lat'),
                ('stlo', 'lon'),
                ('stel', 'elev')]

    sitedict = {}
    for hdr, col in sac_site:
        val = header.get(hdr, None)
        sitedict[col] = val if val != SACDEFAULT[hdr] else None

    # clean up
    try:
        sitedict['elev'] /= 1000.0
    except (TypeError, KeyError):
        # no 'elev'
        pass

    sitedict = _cast_float(sitedict, ['lat', 'lon', 'elev'])
    sitedict = _clean_str(sitedict, ['sta'])

    sitedict['sta'] = sitedict['sta'].strip()[:6]

    return [sitedict] or []


def sachdr2arraysite(header):
    """
    Provide a SAC header dictionary, get a site table dictionary.
    This is a slight modification of sachdr2site that handles deast
    & dnorth being in SAC's user7 & user8 header fields (a poor
    but documented standard that is the de facto appoach
    - see SAC's BBFK or ARRAYMAP or BRTT's db2sac in Antelope).

    Note that user7 & user8 are expected to be in km (consistent
    with SAC expectations).

    """
    sac_site = [('kstnm', 'sta'),
                ('stla', 'lat'),
                ('stlo', 'lon'),
                ('stel', 'elev'),
                ('user7','deast'),
                ('user8','dnorth')]

    sitedict = {}
    for hdr, col in sac_site:
        val = header.get(hdr, None)
        sitedict[col] = val if val != SACDEFAULT[hdr] else None

    # clean up
    try:
        sitedict['elev'] /= 1000.0
    except (TypeError, KeyError):
        #no 'elev'
        pass

    sitedict = _cast_float(sitedict, ['lat', 'lon', 'elev', 'deast', 'dnorth'])
    sitedict = _clean_str(sitedict, ['sta'])

    sitedict['sta'] = sitedict['sta'].strip()[:6]

    return [sitedict] or []


def sachdr2sitechan(header):
    """
    Provide a sac header dictionary, get a sitechan table dictionary.

    """
    sac_sitechan = [('kstnm', 'sta'),
                    ('kcmpnm', 'chan'),
                    ('cmpaz', 'hang'),
                    ('cmpinc', 'vang'),
                    ('stdp', 'edepth')]

    sitechandict = AttribDict()
    for hdr, col in sac_sitechan:
        val = header.get(hdr, None)
        sitechandict[col] = val if val != SACDEFAULT[hdr] else None

    try:
        sitechandict['edepth'] /= 1000.0
    except (TypeError, KeyError):
        # edepth is None or missing
        pass

    sitechandict = _cast_float(sitechandict, ['hang', 'vang', 'edepth'])
    sitechandict = _clean_str(sitechandict, ['sta', 'chan'])
    sitechandict['sta'] = sitechandict['sta'].strip()[:6]

    return [sitechandict] or []


def sachdr2affiliation(header):
    sac_affil = [('knetwk', 'net'),
                 ('kstnm', 'sta')]

    affildict = AttribDict()
    for hdr, col in sac_affil:
        val = header.get(hdr, None)
        affildict[col] = val if val != SACDEFAULT[hdr] else None

    affildict = _clean_str(affildict, ['net', 'sta'])

    affildict['sta'] = affildict['sta'].strip()[:6]

    return [affildict] or []


def sachdr2instrument(header):
    # TODO: investigate hdr['resp0-9'] values
    sac_instr = [('kinst', 'insname'),
                 ('iinst', 'instype'),
                 ('delta', 'samprate')]

    instrdict = AttribDict()
    for hdr, col in sac_instr:
        val = header.get(hdr, None)
        instrdict[col] = val if val != SACDEFAULT[hdr] else None

    # clean up
    try:
        instrdict['samprate'] = int(round(1.0 / instrdict['samprate'], 0))
    except (TypeError, KeyError):
        pass

    return [instrdict] or []


def sachdr2origin(header):
    """
    Provide a sac header dictionary, get a filled origin table dictionary.
    A few things:
    1) If sac reference time isn't event-based, origin time is unknown
    2) magnitude is taken first from hdr['mag'], hdr['imagtyp'] if defined,
       then replaced by hdr['user0'],hdr['kuser0']
    3) sac moment, duration, and user-defined magnitude headers aren't
       put into the origin table
    4) origin.auth is taken first from hdr['imagsrc'], then replaced by
       hdr['kuser1'] if defined.  the imagsrc->auth translations are:

     * INEIC -> ISC:NEIC
     * IPDE -> PDE
     * IISC -> ISC
     * IBRK -> ISC:BERK
     * IUSGS, ICALTECH, ILLNL, IEVLOC, IJSOP, IUSER, IUNKNOWN -> unchanged

    """
    # simple SAC translations
    sac_origin = [('evla', 'lat'),
                  ('evlo', 'lon'),
                  ('norid', 'orid'),
                  ('nevid', 'evid'),
                  ('ievreg', 'grn'),
                  ('evdp', 'depth')]

    origindict = AttribDict()
    for hdr, col in sac_origin:
        val = header.get(hdr, None)
        origindict[col] = val if val != SACDEFAULT[hdr] else None

    # depth
    try:
        origindict['depth'] /= 1000.0
    except (TypeError, KeyError):
        # evdp is None or mising
        pass

    # etype translations
    edict = {37: 'en', 38: 'ex', 39: 'ex', 40: 'qt', 41: 'qt', 42: 'qt',
             43: 'ec', 72: 'me', 73: 'me', 74: 'me', 75: 'me', 76: 'mb',
             77: 'qt', 78: 'qt', 79: 'qt', 80: 'ex', 81: 'ex', 82: 'en',
             83: 'mc'}
    try:
        origindict['etype'] = edict[header['ievtype']]
    except (TypeError, KeyError):
        # ievtyp is None, or not a key in edict
        pass

    # 1:
    try:
        t = get_sac_reftime(header)
        if header['iztype'] == 11:
            # reference time is an origin time
            o = header.get('o', None)
            o = o if (o != SACDEFAULT['o']) else 0.0

            origindict['time'] = t.timestamp - o
            origindict['jdate'] = int((t - o).strftime('%Y%j'))
    except (ValueError, KeyError):
        # no trace.stats.sac, no iztype
        pass

    # 2: magnitude
    magdict = {52: 'mb', 53: 'ms', 54: 'ml'}
    try:
        origindict[magdict[header['imagtyp']]] = header['mag']
    except (ValueError, KeyError):
        # imagtyp is None or not a key in magdict
        pass

    # is kuser0 is a recognized magnitude type, overwrite mag
    # XXX: this is a LANL wfdisc2sac thing
    try:
        magtype = header['kuser0'].strip()
        if magtype in list(magdict.values()):
            origindict[magtype] = header['user0']
    except (KeyError, ValueError):
        # kuser0 is None
        pass

    # 3: origin author
    authdict = {58: 'ISC:NEIC', 61: 'PDE', 62: 'ISC', 63: 'REB-ICD',
                64: 'IUSGS', 65: 'ISC:BERK', 66: 'ICALTECH', 67: 'ILLNL',
                68: 'IEVLOC', 69: 'IJSOP', 70: 'IUSER', 71: 'IUNKNOWN'}
    try:
        origindict['auth'] = authdict[header['imagsrc']]
    except (KeyError, ValueError):
        # imagsrc not in authdict (i.e. sac default value)
        pass

    # XXX: this is LANL wfdisc2sac thing.  maybe turn it off?
    if header.get('kuser1'):
        origindict['auth'] = header['kuser1']

    origindict = _cast_float(origindict, ['lat', 'lon', 'depth'])
    origindict = _cast_int(origindict, ['evid', 'orid'])

    return [origindict] or []


def sachdr2event(header):
    sac_event = [('nevid', 'evid'),
                 ('kevnm', 'evname')]

    eventdict = AttribDict()
    for hdr, col in sac_event:
        val = header.get(hdr, None)
        eventdict[col] = val if val != SACDEFAULT[hdr] else None

    eventdict = _cast_int(eventdict, ['evid'])

    return [eventdict] or []


def sachdr2assoc(header, pickmap=None):
    """
    Takes a sac header dictionary, and produces a list of up to 10
    Assoc instances. Header->phase mappings follow SAC2000, i.e.:

    * t0: P
    * t1: Pn
    * t2: Pg
    * t3: S
    * t4: Sn
    * t5: Sg
    * t6: Lg
    * t7: LR
    * t8: Rg
    * t9: pP

    An alternate mapping for some or all picks can be supplied, however,
    as a dictionary of strings in the above form.

    Note: arid values will not be filled in, so do:
    >>> for assoc in kbio.tables['assoc']:
            assoc.arid = lastarid+1
            lastarid += 1

    """
    pick2phase = {'t0': 'P', 't1': 'Pn', 't2': 'Pg', 't3': 'S',
                  't4': 'Sn', 't5': 'Sg', 't6': 'Lg', 't7': 'LR', 't8': 'Rg',
                  't9': 'pP'}

    # overwrite defaults with supplied map
    if pickmap:
        pick2phase.update(pickmap)

    # geographic relations
    # obspy.read tries to calculate these values if lcalca is True and needed
    # header info is there, so we only need to try to if lcalca is False.
    # XXX: I just calculate it if no values are currently filled in.
    sac_assoc = [('az', 'esaz'),
                 ('baz', 'seaz'),
                 ('gcarc', 'delta')]

    assocdict = AttribDict()
    for hdr, col in sac_assoc:
        val = header.get(hdr, None)
        assocdict[col] = val if val != SACDEFAULT[hdr] else None

    # overwrite if any are None
    if not assocdict:
        try:
            delta = geod.locations2degrees(header['stla'], header['stlo'],
                                           header['evla'], header['evlo'])
            m, seaz, esaz = geod.gps2DistAzimuth(header['stla'], header['stlo'],
                                                 header['evla'], header['evlo'])
            assocdict['esaz'] = esaz
            assocdict['seaz'] = seaz
            assocdict['delta'] = delta
        except (ValueError, TypeError):
            # some sac header values are None
            pass

    if header.get('kstnm', None):
        assocdict['sta'] = header['kstnm']

    orid = header.get('norid', None)
    assocdict['orid'] = orid if orid != SACDEFAULT['norid'] else None

    # now, do the phase arrival mappings
    # for each pick in hdr, make a separate dictionary containing assocdict plus
    # the new phase info.
    assocs = []
    for key in pick2phase:
        kkey = 'k' + key
        # if there's a value in t[0-9]
        if header.get(key, None) not in (SACDEFAULT[key], None):
            # if the phase name kt[0-9] is null
            if header[kkey] == SACDEFAULT[kkey]:
                # take it from the map
                iassoc = {'phase': pick2phase[key]}
            else:
                # take it directly
                iassoc = {'phase': header[kkey]}

            iassoc.update(assocdict)
            assocs.append(iassoc)

    return assocs


def sachdr2arrival(header, pickmap=None):
    """Similar to sachdr2assoc, but produces a list of up to 10 Arrival
    dictionaries.  Same header->phase mapping applies, unless otherwise stated.

    """
    # puts t[0-9] times into arrival.time if they're not null
    # puts corresponding kt[0-9] phase name into arrival.iphase
    # if a kt[0-9] phase name is null and its t[0-9] values isn't,
    # phase names are pulled from the pick2phase dictionary
    pick2phase = {'t0': 'P', 't1': 'Pn', 't2': 'Pg', 't3': 'S',
                  't4': 'Sn', 't5': 'Sg', 't6': 'Lg', 't7': 'LR', 't8': 'Rg',
                  't9': 'pP'}

    if pickmap:
        pick2phase.update(pickmap)

    # simple translations
    arrivaldict = AttribDict()
    if header.get('kstnm', None) not in (SACDEFAULT['kstnm'], None):
        arrivaldict['sta'] = header['kstnm']
    if header.get('kcmpnm', None) not in (SACDEFAULT['kcmpnm'], None):
        arrivaldict['chan'] = header['kcmpnm']

    # phases and arrival times
    t0 = get_sac_reftime(header)
    arrivals = []
    for key in pick2phase:
        kkey = 'k' + key
        # if there's a value in t[0-9]
        if header.get('key', None) not in (SACDEFAULT[key], None):
            # TODO: This seems broken...t isn't defined yet
            itime = t + header[key]
            iarrival = {'time': itime.timestamp,
                        'jdate': int(itime.strftime('%Y%j'))}
            # if the phase name kt[0-9] is null
            if header[kkey] == SACDEFAULT[kkey]:
                # take it from the pick2phase map
                iarrival['iphase'] = pick2phase[key]
            else:
                # take it directly
                iarrival['iphase'] = header[kkey]

            iarrival.update(arrivaldict)
            # TODO: wtf is iassoc??
            arrivals.append(iassoc)

    return arrivals


def sachdr2wfdisc(header):
    """Produce wfdisc kbcore table instance from sac header dictionary.
    Clearly this will be a skeleton instance, as the all-important 'dir' and
    'dfile' and 'datatype' must be filled in later.

    """
    t0 = get_sac_reftime(header)
    b = header.get('b', None)
    b = b if (b != SACDEFAULT['b']) else 0.0
    starttime = t0 + b
    e = header.get('e', None)
    e = e if (e != SACDEFAULT['e']) else 0.0
    endtime = t0 + e

    wfdict = AttribDict()
    wfdict['nsamp'] = int(header.get('npts', None))
    wfdict['time'] = starttime.timestamp
    wfdict['endtime'] = endtime.timestamp
    wfdict['jdate'] = int(starttime.strftime('%Y%j'))

    wfdict['samprate'] = int(round(1.0 / header['delta']))

    kstnm = header.get('kstnm', None)
    if kstnm not in (SACDEFAULT['kstnm'], None):
        wfdict['sta'] = kstnm.strip()[:6]

    kcmpnm = header.get('kcmpnm', None)
    if kcmpnm not in (SACDEFAULT['kcmpnm'], None):
        wfdict['chan'] = kcmpnm.strip()[:8]

    scale = header.get('scale', None)
    if scale not in (SACDEFAULT['scale'], None):
        wfdict['calib'] = float(scale)

    nwfid = header.get('nwfid', None)
    if nwfid not in (SACDEFAULT['nwfid'], None):
        wfdict['wfid'] = nwfid

    wfdict['foff'] = 632

    if sys.byteorder == 'little':
        wfdict['datatype'] = 'f4'
    else:
        wfdict['datatype'] = 't4'

    return [wfdict] or []


def sachdr2tables(header, tables=None):
    """
    Scrape SAC header dictionary into database table dictionaries.

    Parameters
    ----------
    header : dict
        SAC header
    tables : list/tuple of strings, optional
        Table name strings to return.
        Default, ['affiliation', 'arrival', 'assoc', 'event', 'instrument',
        'origin', 'site', 'sitechan', 'wfdisc']

    Returns
    -------
    dict
        Dictionary of lists of table dictionaries.  If only default
        values are found for a table, it is omitted.

    Notes
    -----
    Some things must be checked or filled in before adding to a database:
    - affiliation.time, endtime
    - arrival.arid
    - assoc.orid, arid
    - event.prefor, evid
    - instrument.inid, dir, dfile, rsptype
    - site.ondate
    - sitechan.ondate, chanid
    - wfdisc.dir, dfile, foff, datetype, wfid

    """
    fns = {'affiliation': sachdr2affiliation,
           'arrival': sachdr2arrival,
           'assoc': sachdr2assoc,
           'event': sachdr2event,
           'instrument': sachdr2instrument,
           'origin': sachdr2origin,
           'site': sachdr2site,
           'sitechan': sachdr2sitechan,
           'wfdisc': sachdr2wfdisc}

    if tables is None:
        tables = list(fns.keys())

#     for key in header:
#         if key.startswith('k'):
#             try:
#                 header[key] = header[key].strip()[:6]

    # t = AttribDict()
    t = {}
    for table in tables:
        try:
            itab = fns[table](header)
        except KeyError:
            itab = []

        if itab:
            t[table] = itab

    # t[table] doesn't exist if table's function didn't return anything

    return t


# ----------------- CONVERT TABLES TO SAC HEADER DICTIONARY ---------------#
# TODO: make these functions able to gracefully handle None values as inputs
def site2sachdr(s):
    """
    Accepts a fielded site table row and returns a dictionary of corresponding
    sac header field/value pairs.

    June 13, 2016 - added deast/dnorth output to SAC user7/8 header fields
    """
    keymap = {'stel': 'elev', 'stal': 'lat', 'stlo': 'lon',
        'user7': 'deast', 'user8': 'dnorth'}
    hdr = _buildhdr(keymap, s)
    return hdr


def sitechan2sachdr(sc):
    keymap = {'cmpaz': 'hang', 'cmpinc': 'vang'}
    hdr = _buildhdr(keymap, sc)
    return hdr


def affiliation2sachdr(af):
    keymap = {'network': 'net'}
    hdr = _buildhdr(keymap, af)
    return hdr


def instrument2sachdr(ins):
    return {}


def origin2sachdr(o):
    """
    Accepts a fielded origin table record and produces a dictionary of
    corresponding sac header field/value pairs.
    """
    keymap = {'evdp': 'depth', 'evla': 'lat', 'evlo': 'lon', 'kuser1': 'auth',
              'nevid': 'evid', 'norid': 'orid', 'user0': 'mb'}
    hdr = _buildhdr(keymap, o)
    return hdr


def event2sachdr(evt):
    return {}


def assoc2sachdr(asc):
    return {}


def arrival2sachdr(ar):
    return {}


def wfdisc2sachdr(wf):
    pass


# functions that accept a table, return a dictionary of sac header values
# the order of this dictionary matters
KB2SAC = OrderedDict({'site': site2sachdr,
                      'sitechan': sitechan2sachdr,
                      'wfdisc': wfdisc2sachdr,
                      'affiliation': affiliation2sachdr,
                      'instrument': instrument2sachdr,
                      'origin': origin2sachdr,
                      'event': event2sachdr,
                      'assoc': assoc2sachdr,
                      'arrival': arrival2sachdr})


def tables2sachdr(tables):
    """Returns a sac header dictionary, including default values, from
    current table instances.  SAC reference time is, in order of availability,
    origin time (origin.time), first sample time (wfdisc.time).

    """

    hdr = SACDEFAULT.copy()
    for table, tabfun in KB2SAC.items():
        hdr.update(tabfun(tables.get(table, None)))

    return hdr



# -*- coding: utf-8 -*-
"""
Common utility functions.

"""
import logging
import math
from getpass import getpass
import warnings
import functools
import inspect
from importlib import import_module

import numpy as np
import sqlalchemy as sa
from sqlalchemy.orm import Session
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.exc import IntegrityError
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.orm.exc import UnmappedInstanceError

import obspy.geodetics as geod
from obspy.core import AttribDict
from obspy.taup import TauPyModel

from pisces.schema.util import PiscesMeta


def deprecated(instructions):
    """
    Flags a method as deprecated.

    Parameters
    ----------
    instructions : str
        A human-friendly string of instructions, such as:
        'Please migrate to add_proxy() ASAP.'

    References
    ----------
    https://gist.github.com/kgriffs/8202106

    """
    def decorator(func):
        '''This is a decorator which can be used to mark functions
        as deprecated. It will result in a warning being emitted
        when the function is used.'''
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            message = 'Call to deprecated function {}. {}'.format(
                func.__name__,
                instructions)

            frame = inspect.currentframe().f_back

            warnings.warn_explicit(message,
                                   category=DeprecationWarning,
                                   filename=inspect.getfile(frame.f_code),
                                   lineno=frame.f_lineno)

            return func(*args, **kwargs)

        return wrapper

    return decorator

TURNOFFWARNINGSMSG = """Warnings can be turned off with:
import warnings
warnings.filterwarnings('ignore', category=DeprecationWarning
"""

def db_connect(*args, **kwargs):
    """
    Connect to your database.

    Parameters
    ----------
    backend : string
        One of the SQLAlchemy connection strings from
        http://docs.sqlalchemy.org/en/rel_0_7/core/engines.html#database-urls
    user : string, optional
        Not required for sqlite.
    passwd : string, optional
        Not needed for sqlite. Prompted if needed and not provided.
    server : string, optional
        Database host server.
    port : string or integer, optional
        Port on remote server.
    instance : string, optional
        The database instance.  For sqlite, this is the file name.
    conn : string, optional
        A fully-formed SQLAlchemy style connection string.

    Returns
    -------
    session : bound SQLAlchemy Session instance

    Examples
    --------
    1. Connect to a local sqlite database file:

    >>> meta, session = db_connect(conn='sqlite:///mydb.sqlite')
    #or
    >>> meta, session = db_connect(backend='sqlite', instance='mydb.sqlite')

    Notes
    -----
    For connection string format, see:
    http://docs.sqlalchemy.org/en/rel_0_8/core/engines.html

    """
    #TODO: take advantage of sqlalchemy.engine.url.URL
    #XXX: is not protected against using args and kwargs
    if len(args) == 1:
        kwargs['conn'] = args[0]

    if kwargs.get('conn'):
        conn = kwargs.get('conn')
    else:
        backend = kwargs.get('backend')
        user = kwargs.get('user', '')
        psswd = kwargs.get('passwd', '')
        server = kwargs.get('server', '')
        port = kwargs.get('port', '')
        instance = kwargs.get('instance', '')

        if backend is 'sqlite':
            userpsswd = ''
        else:
            if user and not psswd:
                psswd = getpass("Enter password for {0}: ".format(user))
                userpsswd = ':'.join([user, psswd])
            elif psswd and not user:
                user = getpass("Enter username for given password: ")
                userpsswd = ':'.join([user, psswd])
            elif user and psswd:
                userpsswd = ':'.join([user, psswd])
            else:
                userpsswd = ':'

        if server:
            serverport = '@' + server
            if port:
                serverport += ':' + str(port)
        else:
            serverport = ''

        conn = "{0}://{1}{2}/{3}".format(backend, userpsswd, serverport, instance)

    engine = sa.create_engine(conn)
    session = Session(bind=engine)

    return session

# TODO: get rid of two different ways to connect to databases
def url_connect(url):
    """
    Connect to a database using an RFC-1738 compliant URL, like sqlalchemy's
    create_engine, prompting for a password if a username is supplied.

    Parameters
    ----------
    url : string
        A fully-formed SQLAlchemy style connection string.
        See http://docs.sqlalchemy.org/en/latest/core/engines.html#database-urls

    Returns
    -------
    session : bound SQLAlchemy Session instance

    Examples
    --------
    SQLite database file, local:
    >>> url_connect('sqlite:///local/path/to/mydb.sqlite')

    SQLite database file, full path:
    >>> url_connect('sqlite:////full/path/to/mydb.sqlite')

    Remote Oracle, OS-authenticated (no user or password needs to be specified)
    >>> url_connect('oracle://dbserver.lanl.gov:8080/mydb')

    Remote Oracle, password-authenticated (specify user, prompted for password)
    >>> url_connect('oracle://scott@dbserver.lanl.gov:8080/mydb')
    Enter password for scott:

    Remote Oracle, password-authenticated (password specified)
    >>> url_connect('oracle://scott:tiger@dbserver.lanl.gov:8080/mydb')

    """
    this_url = sa.engine.url.make_url(url)
    if this_url.username and not this_url.password:
        this_url.password = getpass("Enter password for {0}: ".format(this_url.username))

    e = sa.create_engine(this_url)

    session = Session(bind=e)

    return session

# TODO: rename this to "load_table", and make it work on a single table
@deprecated('get_tables is deprecated. It will be load_tables in pisces.crud in next version. ' + TURNOFFWARNINGSMSG)
def get_tables(bind, fulltablenames, metadata=None, primary_keys=None,
               base=None):
    """
    Reflect/load an arbitrary database table as a mapped class.

    This is a shortcut for SQLAlchemy's declarative mapping using __table__.
    See http://docs.sqlalchemy.org/en/rel_0_9/orm/extensions/declarative.html#using-a-hybrid-approach-with-table.

    Parameters
    ----------
    bind : sqlalchemy.engine.base.Engine instance
        Engine pointing to the target database.
    fulltables : list of strings
        Of the form ['owner1.tablename1', 'owner2.tablename2', ...]
        Leave out 'owner.' if database doesn't use owners (sqlite, etc...)
    metadata : sqlalchemy.MetaData, optional
        MetaData into which reflected Tables go. If not supplied, a new one
        is created, accessible from MyTable.metadata on one of the loaded
        tables.
    primary_keys : dict, optional
        Tablename, primary key list pairs of the form,
        {'owner1.tablename1': ['primary_key1', 'primary_key2']}
        These are required if the table is a view or has no primary key.
    base : sqlalchemy.ext.declarative.api.DeclarativeMeta, optional
        The declarative base the from which loaded table classes will inherit.
        The info dictionary of loaded Columns will be updated from those in
        the base.  These are used to generate default values and string
        representations.  Import from pisces.schema.css3, or extensions thereof.
        Default, sqlalchemy.ext.declarative.api.DeclarativeMeta.

    Returns
    -------
    list
        Corresponding list of ORM table classes mapped to reflected tables,
        Can be used for querying or making row instances.

    Raises
    ------
    sqlalchemy.exc.NoSuchTableError : Table doesn't exist.
    sqlalchemy.exc.InvalidRequestError : Table already loaded in metadata.
    sqlalchemy.exc.ArgumentError : Table has no primary key(s).

    Notes
    -----
    In SQLAlchemy, a database account/owner is generally used with the "schema"
    keyword argument.

    For core tables in a Pisces schema, this function isn't recommended.
    Instead, subclass from the known abstract table.

    Examples
    --------
    # for unknown table
    >>> import pisces.schema.css3 as css
    >>> RandomTable = get_tables(engine, ['randomtable'])

    # for a known/prototype table
    >>> class Site(css.Site):
            __tablename__ = 'myaccount.my_site_tablename'

    """
    if not metadata:
        metadata = sa.MetaData()
        if not bind:
            raise ValueError("Must provide bound metadata or bind.")
    # we have metadata
    if not bind:
        bind = metadata.bind

    ORMBase = base if base else declarative_base(metaclass=PiscesMeta,
                                                 constructor=None,
                                                 metadata=metadata)
    colinfo = getattr(base, '_column_info_registry', {})
    parents = (ORMBase,)


    outTables = []
    for fulltable in fulltablenames:
        try:
            owner, tablename = fulltable.split('.')
        except ValueError:
            # no owner given
            owner, tablename = None, fulltable

        # reflect the table
        itable = sa.Table(tablename, metadata, autoload=True,
                          autoload_with=bind, schema=owner)

        # update reflected table with known schema column info
        if colinfo:
            for col in itable.columns:
                col.info.update(colinfo.get(col.name, {}))

        dct = {'__table__': itable}
        # put any desired __table_args__: {} here

        # no primary key, can't map. spoof primary key mapping from inputs
        if primary_keys and fulltable in primary_keys:
            dct['__mapper_args__'] = {'primary_key': [getattr(itable.c, key) for key in primary_keys[fulltable]]}

        ORMTable = type(tablename.capitalize(), parents, dct)

        outTables.append(ORMTable)

    return outTables


# TODO: merge get_tables and make_tables?
@deprecated('make_table will be moved to pisces.crud in next version .' + TURNOFFWARNINGSMSG)
def make_table(fulltablename, prototype):
    """
    Create a new ORM class/model on-the-fly from a prototype.

    Parameters
    ----------
    fulltablename : str
        Schema-qualified name of the database table, like 'owner.tablename'
        or just 'tablename'.  The resulting classname will be the capitalized
        tablename, like 'Tablename'.
    prototype : sqlalchemy abstract ORM class
        The prototype table class. pisces.schema.css.Site, for example.

    Notes
    -----
    It's better to declare classes in an external module, and import them.
    SQLAlchemy doesn't let you use the same table names twice, so on-the-fly
    class creation and naming is risky:

    1. You can't use make_tables again if you accidentally overwrite the
       variable you used to hold the class you created.
    2. You can't use make_tables again if you import something from a
       script/module where make_tables was used with the same table name.

    """
    try:
        owner, tablename = fulltablename.split('.')
    except ValueError:
        owner, tablename = None, fulltablename

    parents = (prototype,)
    if owner:
        OwnerBase = declarative_base(metadata=sa.MetaData(schema=owner))
        parents = (OwnerBase, prototype)
    else:
        parents = (prototype,)

    dct = {'__tablename__': tablename}

    return type(tablename.capitalize(), parents, dct)


def make_same_size(lat1, lon1, lat2, lon2):
    """
    Returns numpy arrays the same size as longest inputs.
    assume: lat1/lon1 are same size and lat2/lon2 are same size
    assume: the smaller of the sizes is a scalar

    """
    #TODO: EAFP
    lon1 = np.array(lon1)
    lat1 = np.array(lat1)
    lon2 = np.array(lon2)
    lat2 = np.array(lat2)
    #assume: lat1/lon1 are same size and lat2/lon2 are same size
    #assume: the smaller of the sizes is a scalar
    N1 = lon1.size
    N2 = lon2.size
    if N1 > N2:
        lon2 = lon2.repeat(N1)
        lat2 = lat2.repeat(N1)
    elif N2 > N1:
        lon1 = lon1.repeat(N2)
        lat1 = lat1.repeat(N2)

    return lat1, lon1, lat2, lon2


def gen_id(i=0):
    """
    Produce a generator for sequential integer id values.

    Examples
    --------
    >>> lastorid = 7
    >>> orid = gen_id(lastorid)
    >>> orid.next()
    8
    >>> orid.next()
    9

    Generate more than one at a time:

    >>> orid, arid, wfid = (gen_id() for id in ['orid', 'arid', 'wfid'])
    >>> orid.next(), arid.next()
    (1, 1)

    Dictionary of id generators for desired ids, starting where they left off.
    ids not in Lastid will be missing

    >>> ids = session.query(Lastid).filter(Lastid.keyname.in_(['orid','arid']).all()
    >>> last = dict([(id.keyname, gen_id(id.keyvalue)) for id in ids])
    >>> last['orid'].next()
    8820005

    """
    while 1:
        i += 1
        yield i

ak135 = TauPyModel(model='ak135')

def travel_times(ref, deg=None, km=None, depth=0.):
    """
    Get *approximate* relative travel time(s).

    Parameters
    ----------
    ref : list or tuple of strings and/or floats
        Reference phase names or horizontal velocities [km/sec].
    deg : float, optional
        Degrees of arc between two points of interest (spherical earth).
    km : float, optional
        Horizontal kilometers between two points of interest (spherical earth).
    depth : float, optional. default, 0.
        Depth (positive down) of event, in kilometers.

    Returns
    -------
    numpy.ndarray
        Relative times, in seconds, same length as "ref". NaN if requested time
        is undefined.

    Examples
    --------
    Get relative P arrival and 2.7 km/sec surface wave arrival at 35 degrees
    distance.
    >>> times = travel_times(['P', 2.7], deg=35.0)
    To get absolute window, add the origin time like:
    >>> w1, w2 = times + epoch_origin_time

    Notes
    -----
    Either deg or km must be indicated.
    The user is responsible for adding/subtracting time (such as origin
    time, pre-window noise time, etc.) from those predicted in order to define
    a window.
    Phase travel times use ak135.

    """
    times = np.zeros(len(ref), dtype='float')
    tt = None
    for i, iref in enumerate(ref):
        if isinstance(iref, str):
            # phase time requested
            if not tt:
                if not deg:
                    deg = geod.kilometer2degrees(km)
                tt = ak135.get_travel_times(depth, deg)
            try:
                idx = [ph.name for ph in tt].index(iref)
                itt = [ph.time for ph in tt][idx]
            except ValueError:
                # phase not found
                itt = None
        else:
            # horizontal velocity
            if not km:
                km = deg*(2*math.pi/360.0)*6371.0
            itt = km/iref
        times[i] = itt

    return times


def add_rows(session, rows, recurse=False):
    """Handle common errors with logging in SQLAlchemy add_all.

    Tries to add in bulk.  Failing that, it will rollback and optionally try
    to add one at a time.

    Parameters
    ----------
    session : sqlalchemy.orm.Session
    rows : list
        Mapped table instances.
    recurse : bool, optional
        After failure, try to add records individually.

    Returns
    -------
    num : int
        Number of objects added.  0 if none.
    e : exception or None

    """
    e = None
    num = 0
    try:
        session.add_all(rows)
        session.commit()
        num = len(rows)
    except (ProgrammingError, UnmappedInstanceError) as e:
        # IntegrityError: duplicate row(s)
        # ProgrammingError: string encoding problem
        # UnmappedInstanceError: tried to add something like a list or None
        session.rollback()
        logging.warning(str(e))
    except IntegrityError:
        print(str(e))
        session.rollback()
    finally:
        # always executed
        if e and recurse:
            # if an exception was thrown and recursion was requested
            for row in rows:
                i, e = add_rows(session, [row], recurse=False)
                num += i

    return num, e


def get_lastids(session, Lastid, keynames=None, expunge=True, create=False):
    """
    Load or create Lastid instances into a convenient and readable
    attribute-based dictionary.

    Parameters
    ----------
    session : sqlalchemy.orm.Session instance
    Lastid : Lastid table class
    ids : list or tuple of strings
        Lastid.keyname values to load.
    expunge : bool
        If True, expunge loaded ids from the session.  This frees you
        to modify them without affecting the database from which they
        were loaded.  In this case, you'll have to add them back into a
        session and commit them for their changes to be reflected on the
        database.
    create : bool
        If True, create ids that don't already exist.

    Returns
    -------
    last : obspy.core.AttribDict
        Attribute-based dictionary of all lastids.


    Examples
    --------
    Get and set lastid values directly by name or by attribute.
    >>> last = get_lastids(session, Lastid, ['orid', 'arid'])
    >>> last.orid, last['orid']
    Lastid(keyname='orid'), Lastid(keyname='orid')

    Test for their existence by name.
    >>> 'orid' in last
    True

    Use the Lastid's 'next' generator behavior for readable code
    >>> next(last.orid)
    18
    >>> last.orid.keyvalue
    18

    Update your database when you're done.
    >>> session.add_all(ids.values())
    >>> session.commit()

    """

    last = AttribDict()

    q = session.query(Lastid)
    if keynames is None:
        lastids = q.all()
    else:
        lastids = []
        for keyname in keynames:
            lastid = q.filter(Lastid.keyname == keyname).first()
            if lastid:
                lastids.append(lastid)
            elif create:
                lastid = Lastid(keyname=keyname, keyvalue=0)
                session.add(lastid)
                lastids.append(lastid)

    for lastid in lastids:
        if expunge:
            session.expunge(lastid)
        last[lastid.keyname] = lastid

    return last


def get_options(db, prefix=None):
    '''
    for coretable in CORETABLES:
        table_group.add_argument('--' + coretable.name,
                            default=None,
							metavar='owner.tablename',
							dest=coretable.name)
    '''
    options={'url':'sqlite:///'+db, 'prefix':prefix}

    return options


@deprecated('get_or_create_tables will be moved to pisces.crud in next version .' + TURNOFFWARNINGSMSG)
def get_or_create_tables(session, create=True, **tables):
    """
    Load or create canonical ORM KB Core table classes.

    Parameters
    ----------
    session : sqlalchemy.orm.Session
    create : bool
        If True, create a table object that isn't found.
    tables
        Canonical table name / formatted table name keyword pairs.

    Also accepted are canonical table name keywords with '[owner.]tablename'
    arguments, which will replace any prefix-based core table names.

    Returns
    -------
    tables : dict
        Mapping between canonical table names and SQLA ORM classes.
        e.g. {'origin': MyOrigin, ...}

    """
    # The Plan:
    # 1. For each core table, build or get the table name
    # 2. If it's a vanilla table name, just use a pre-packaged table class
    # 3. If not, try to autoload it.
    # 4. If it doesn't exist, make it from a prototype and create it in the database.

    # TODO: check options for which tables to produce.

    tables = {}
    for coretable in CORETABLES:
        # build the table name
        if options['prefix'] == None:
            fulltablename = coretable.name
        else:
            fulltablename = prefix + coretable.name

        # fulltablename is either an arbitrary string or prefix + core name, but not None

        # put table classes into the tables dictionary
        if fulltablename == coretable.name:
            # it's a vanilla table name. just use a pre-packaged table class instead of making one.
            tables[coretable.name] = coretable.table
        else:
            tables[coretable.name] = make_table(fulltablename, coretable.prototype)

        tables[coretable.name].__table__.create(session.bind, checkfirst=True)

    session.commit()

    return tables


def load_config(config):
    """
    Take dictionary-like object (actual dictionary or ConfigParser section)
    to produce a SQLAlchemy database session and dictionary of database table
    classes.

    Parameters
    ----------
    config : dict-like

    Returns
    -------
    session : SQLAlchemy database session instance
    tables : dict
        Keys are canonical table names, values are SQLAlchemy table classes.

    The config input is a dict-like object of table name keys and class paths.
    This can be a configparser object or a nested dict. The "database" section and "url" key
    are required.  Any additional parameters in the section must be importable
    module names, with optional SQLAlchemy table class names separated by a colon ":".
    If no class name is provided, a class with the capitalized key name is imported.


    Examples
    --------
    The following inputs should be equivalent.

    # pisces.cfg
    [database] # required
    url = oracle://myuser:mypass@dbserver.someplace.gom:1521/dbname # required
    site = mytablemodule:SiteClassName
    wfdisc = mytablemodule:WfdiscClassName
    mytable = custom_tables:MyCustomTableClass
    arrival = othermodule

    >>> config = {'url': 'oracle://myuser:mypass@dbserver.someplace.gom:1521/dbname',
                  'site': 'modulename:SiteClassName',
                  'wfdisc': 'modulename:Wfdisc_raw',
                  'mytable': 'custom_tables:MyCustomTableClass',
                  'arrival': 'othermodule',
                  }
    # or
    >>> import configparser
    >>> config = configparser.ConfigParser()
    >>> config.read('pisces.cfg')
    >>> session, tables = load_config(config['database'])
    >>> tables
    ... {'arrival': othermodule.Arrival,
    ... 'mytable': custom_tables.MyCustomTableClass,
    ... 'site': modulename.SiteClassName,
    ... 'wfdisc': modulename.Wfdisc_raw}

    """
    URL = config.pop('url')
    session = url_connect(URL)

    tables = {}
    for table, classpath in config.items():
        try:
            module_name, class_name = classpath.split(':')
        except ValueError:
            module_name, class_name = classpath, table.capitalize()

        m = import_module(module_name)
        tables[table] = getattr(m, class_name)

    # put it back
    config['url'] = URL

    return session, tables



# Converted to python 3.5.2 on 02/07/17
# By: Jeremy Webster

"""
Common io utilities.

"""


def _buildhdr(keymap, rec):
    """
    Map attributes in rec into a dictionary, with keyname translations applied.

    Raises
    ------
    AttributeError
        rec does not have attributes (is None?)

    """
    # TODO: make keymap optional?
    # TODO: put this in io.util?
    hdr = {}
    for key1, key2 in keymap.items():
        try:
            hdr[key1] = getattr(rec, key2)
        except AttributeError:
            # rec doesn't have attribute or rec is None. skip
            pass

    return hdr


def _map_header(keymap, dold, nulldict=None):
    """
    Returns a dictionary of values from dictionary dold,
    mapped to new key, if provided.

    Parameters
    ----------
    keymap: dict
        The map between old and new dictionary keys.
        Of the form {oldkey: newkey, ...}
    dold: dict
        The values to hand over.
    nulldict: dict, optional
        Dictionary representing null values in source dictionary.
        Matching values will not be transferred.

    Returns: dict

    """
    # TODO: this confusing function should be rewritten  (using defaultdict?)
    dnew = {}
    if nulldict:
        for oldkey, newkey in keymap.items():
            try:
                if dold[oldkey] != nulldict[oldkey]:
                    if isinstance(dold[oldkey], str):
                        dnew[newkey] = dold[oldkey].strip()
                    else:
                        dnew[newkey] = dold[oldkey]
            except KeyError:
                # one of the dictionaries doesn't have a key
                pass
    else:
        for oldkey, newkey in keymap.items():
            try:
                if isinstance(dold[oldkey], str):
                    dnew[newkey] = dold[oldkey].strip()
                else:
                    dnew[newkey] = dold[oldkey]
            except KeyError:
                pass

    return dnew



# coding: utf-8
"""
Pisces is a practical and extensible data management library in Python.  It 
leverages existing widely-used free and open-source technologies, such as SQL 
databases and Python, in order to provide a seismological data management 
solution that: 

1) allows the user to both manage and analyze data with a single easy-to-learn 
    language, Python, 
2) leverages large existing user communities to facilitate adoption and 
    problem solving in code development, and 
3) imposes no expensive or restrictive licensing constraints on users.  

The ultimate goal of Pisces is to allow the user to write code that will not 
eventually have to be abandoned due to different project scales, system 
architectures, or licensing concerns.  

"""

__version__ = '0.3.0'

from pisces.util import db_connect, get_tables, travel_times, make_table
from pisces.schema.util import string_formatter
from pisces.io.trace import wfdisc2trace
from pisces.schema import kbcore
from pisces.io.readwaveform import read_waveform



import sqlalchemy as sa
from sqlalchemy import create_engine, func, or_
from sqlalchemy.orm import Session

import pisces as ps
import pisces.request as req
from pisces.io.trace import wfdisc2trace


class Client(object):
    """
    Basic database request client.

    """
    def __init__(self, **kwargs):
        """ 
        Initialize a database connection. 

        Parameters
        ----------
        backend: string, optional
            One of the SQLAlchemy connection strings from 
            http://docs.sqlalchemy.org/en/rel_0_7/core/engines.html#database-urls
        user: string, optional
            Not required for sqlite.
        psswd: string, optional
            Not needed for sqlite. Prompted if needed and not provided.
        server: string, optional
            Database host server.
        port: string or integer, optional
            Port on remote server.
        instance: string, optional
            For sqlite, this is the database file name.
        conn: string, optional
            A fully-formed SQLAlchemy style connection string.
        session : sqlalchemy.orm.Session
            An existing session instance.

        Examples
        --------
        >>> from pisces.client import Client
        >>> client = Client(user='scott', backend='oracle', 
                            server='my.server.edu', port=1521, 
                            instance='mydb')

        """
        if kwargs.get('session', None):
            session = session
        else:
            session = ps.db_connect(**kwargs)
        self.metadata = sa.MetaData(session.bind)
        self.session = session

        self.tables = {}

        # do ID management in this class also?

    def load_tables(self, **tables):
        """
        Load core tables.

        Tables are given as key=value pairs as in:
        load_tables(wfdisc='global.wfdisc_raw', origin='location.origin')

        Raises
        ------
        sqlalchemy.exc.NoSuchTableError
            Table doesn't exist.

        """
        # TODO: add the tables as attributes?
        # TODO: if isinstance(table, DeclarativeBase): just set it as attribute
        # XXX: fails for no primary key.  use get_tables syntax.
        for coretable, tablename in tables.iteritems():
            self.tables[coretable] = ps.get_tables(self.session.bind, 
                                                   [tablename], 
                                                   metadata=self.metadata)[0]
        
    def get_events(self, region=None, deg=None, km=None, swath=None, mag=None, 
            depth=None, etype=None, orids=None, evids=None, prefor=False):
        """ Get Origin table rows.

        For further explanation, see: pisces.request.get_events

        """
        #origin = getattr(self, 'origin', None)
        origin = self.tables.get('origin')
        origin = self.tables.get('event', None)

        recs = req.get_events(self.session, origin, event=event, region=region, 
                              deg=deg, km=km, swath=swath, mag=mag, depth=depth,
                              etime=etime, orids=orids, evids=evids, 
                              prefor=prefor)
        return recs

    def get_stations(self, stations=None, channels=None, nets=None, loc=None, 
            region=None, deg=None, km=None, swath=None, stime=None):
        """
        Get Site table rows.

        For further explanation, see: pisces.request.get_stations

        """
        site = self.tables.get('site')
        sitechan = self.tables.get('sitechan', None)
        affiliation = self.tables.get('affiliation', None)

        recs = req.get_stations(self.session, site, sitechan=sitechan, 
                affiliation=affiliation, stations=stations, channels=channels,
                nets=nets, loc=loc, region=region, deg=deg, km=km, swath=swath,
                stime=stime)

        return recs


    def get_arrivals(self, stations=None, channels=None, atime=None, 
            phases=None, arids=None, orids=None, auth=None):
        """
        Request database arrivals.
        
        For further explanation, see: pisces.request.get_arrivals

        """
        arrival = self.tables.get('arrival')
        assoc = self.tables.get('assoc', None)

        recs = get_arrivals(self.session, arrival, assoc=assoc, 
                            stations=stations, channels=channels, atime=atime, 
                            phases=phases, arids=arids, orids=orids, auth=auth)

        return recs


    def get_waveforms(self, station=None, channel=None, starttime=None, 
            endtime=None, wfids=None):
        """
        Get waveforms. 

        For further explanation, see: pisces.request.get_waveforms

        """
        wfdisc = self.tables.get('wfdisc')
        st = get_waveforms(self.session, wfdisc, station=station, 
                           channel=channel, starttime=starttime, 
                           endtime=endtime, wfids=wfids)

        return st



"""
Command-line interface to common Pisces functionality.

This module imports functions from elsewhere in Pisces, wraps them inside local
functions to perform certain common tasks, and uses Click to provide a
command-line interface.  These interfaces functions are organized under a single
nested interface, which is then exported to the command-line using setuptools
entry_points.


DEVELOPERS:
To add a subcommand to the main command, do the following:
1. Add a module file in pisces/commands/ for your implementation
2. Program your functionality into a single function in your modules, and
   import it here.  The module should really just be an ordering of Pisces
   library functions/classes.  The Pisces library functions/classes work with
   native Python objects. The commands submodules do the business of converting
   command-line arguments to a useful form for those library functions.
3. Write a wrapper function for your implementation function here,
   and decorate it with Click.  The docstring for this wrapper function is the
   one that is exposed at the command-line.
   @cli.command('command_name') adds a new "pisces command_name"
   @cli.group('subcommand_name') adds a new "pisces subcommand_name" subcommand

"""

import click

from pisces import __version__
from pisces.util import url_connect
from pisces.commands import sac2db

def split_commas(ctx, param, value):
    """
    Convert from a comma-separated list to a true list.

    """
    # ctx, param, value is the required calling signature for a Click callback
    try:
        values = value.split(',')
    except AttributeError:
        # values is None
        values = None

    return values


# ------------------------------- MAIN ----------------------------------------
# This is the main function/group.  It does nothing except provide a top-level
# --help and serve as a single point of invocation for other subcommands, like
# the way "git" works.  Also, since click.option doesn't support help=, we
# document common arguments, like URL, in the main function.

# "group" means that the command/function can take sub-commands
@click.group(context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(__version__, '-V', '--version')
def cli():
    """
    Pisces command-line interface.

    Commonly-used functionality is exposed as subcommands of this top-level
    function. See "Commands" for things you can do.  Common ARGS are described
    below.

    \b
    Arguments:
      DB      SQLAlchemy-compatible database URI string.
              e.g. sqlite:///localdb.sqlite
                   oracle://user[:password]@server:port/database
                      (leave out password blank for prompt)
              May also be defined as an environmental variable PISCESDB.
        

    """
    pass


# ------------------------------- INIT ----------------------------------------
@cli.command('create')
@click.argument('DB')
def create_command(**kwargs):
    """
    Create core tables.

    Not yet implemented.

    """
    print("create: {}".format(kwargs))


# ------------------------------- DROP ----------------------------------------
@cli.command('drop')
@click.argument('DB')
def drop_command(**kwargs):
    """
    Drop core tables.

    Not yet implemented.

    """
    print("drop: {}".format(kwargs))


# ------------------------------- SAC2DB --------------------------------------
@cli.command('sac2db')
@click.argument('DB', envvar='PISCESDB')
@click.option('-t', '--tables',
              help=("Comma-separated (no spaces), list of tables to create.  "
                    "Default is all core tables with standard names."),
              metavar="owner.tablename[,...]",
              callback=split_commas)
@click.option('-p', '--prefix', default="",
              help=("Target tables using 'account.prefix naming.  "
                    "e.g. myaccount.test_ will target tables like "
                    "myaccount.test_origin, myaccount.test_sitechan."))
@click.option('-A', '--absolute_paths', is_flag=True,
              help=("If set, write database 'dir' directory entries as"
                    " absolute paths, not relative."))
@click.option('--bbfk', is_flag=True,
              help=("If set, get site.deast and dnorth from SAC user7 & user8"
                    " header fields."))
@click.option('-l', '--file_list', type=click.File('r'),
              help="A list file, one file name per line.")
@click.argument('files', nargs=-1, type=click.Path())
def sac2db_command(**kwargs):
    """
    Scrape SAC files into database tables.

    SAC files may be used to produce the following tables: Wfdisc, Site,
    Sitechan, Origin, Event, Arrival, Assoc, Lastid, and Instrument.  Id
    numbering will follow the Lastid table, if one is found, otherwise it will
    start from 1.

    Examples
    --------
    # use standard table names to local test.sqlite file
    pisces sac2db sqlite:///test.sqlite datadir/*.sac

    # prefix all tables in an oracle account with prefix my_, prompt for password
    pisces sac2db --prefix my_ oracle://user@server.domain.com:port/dbname datadir/*.sac

    # if there are too many SAC files for the shell to handle, use a list:
    find datadir -name "*.sac" -print > saclist.txt
    sac2db.py sqlite:///test.sqlite saclist.txt

    """
    # common local functions
    session = url_connect(kwargs['db'])

    # command-specific funtions
    sac2db.main(session=session, **kwargs)


# ------------------------------- MSEED2DB ------------------------------------
@cli.command('mseed2db')
@click.argument('DB')
def mseed2db_command(**kwargs):
    """
    Scrape MSEED files into database tables.

    Not yet implemented.

    """
    print("mseed2db: {}".format(kwargs))


# ------------------------------- QUERY ---------------------------------------
# This is where elementary querying is done, mostly using pisces.request
#
@cli.group('query')
@click.argument('DB')
def query(**kwargs):
    """
    Perform a basic query.

    Not yet implemented.

    """
    print("query: {}".format(kwargs))


@query.command('stations')
@click.argument('DB')
def query_stations(**kwargs):
    """
    Query stations from the site table.

    Not yet implemented.

    """
    print("stations: {}".format(kwargs))

@query.command('events')
@click.argument('DB')
def query_events(**kwargs):
    """
    Query the origin table for events.

    Not yet implemented.

    """
    print("events: {}".format(kwargs))

@query.command('waveforms')
@click.argument('DB')
def query_waveforms(**kwargs):
    """
    Query waveforms from the wfdisc table.

    Not yet implemented.

    """
    print("waveforms: {}".format(kwargs))


if __name__ == '__main__':
    cli()
